from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from pywa.types.others import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.others import (
    Result as _Result,
    User as _User,
    UsersBlockedResult as _UsersBlockedResult,
    UsersUnblockedResult as _UsersUnblockedResult,
    _T,
)

if TYPE_CHECKING:
    from ..client import WhatsApp as WhatsAppAsync


@dataclasses.dataclass(frozen=True, slots=True)
class User(_User):
    """
    Represents a WhatsApp user.

    Attributes:
        wa_id: The WhatsApp ID of the user (The phone number with the country code).
        name: The name of the user (``None`` on :class:`MessageStatus` or when message type is :class:`MessageType.SYSTEM`).
        input: The input of the recipient is only available when sending a message.
    """

    _client: WhatsAppAsync = dataclasses.field(repr=False, hash=False, compare=False)

    async def block(self) -> bool:
        """
        Block the user.

        - Shortcut for :meth:`~pywa_async.client.WhatsApp.block_users` with the user wa_id.

        Returns:
            bool: True if the user was blocked

        Raises:
            BlockUserError: If the user was not blocked
        """
        res = await self._client.block_users((self.wa_id,))
        added = self.wa_id in {u.input for u in res.added_users}
        if not added:
            raise res.errors
        return added

    async def unblock(self) -> bool:
        """
        Unblock the user.

        - Shortcut for :meth:`~pywa_async.client.WhatsApp.unblock_users` with the user wa_id.

        Returns:
            bool: True if the user was unblocked, False otherwise.
        """
        return self.wa_id in {
            u.input
            for u in (await self._client.unblock_users((self.wa_id,))).removed_users
        }

    async def get_call_permissions(self) -> CallPermissions:
        """
        Get the call permissions of the user.

        - Shortcut for :meth:`~pywa.client.WhatsApp.get_call_permissions` with the user wa_id.

        Returns:
            CallPermissions: The call permissions of the user.
        """
        return await self._client.get_call_permissions(wa_id=self.wa_id)


@dataclasses.dataclass(slots=True, frozen=True)
class UsersBlockedResult(_UsersBlockedResult):
    """
    Represents the result of blocking users operation.

    Attributes:
        added_users: The users that were successfully blocked.
        failed_users: The users that failed to be blocked. You can access the .errors attribute in each failure to get the error details.
        errors: The errors that occurred during the operation (if any).
    """

    added_users: tuple[User, ...]


@dataclasses.dataclass(frozen=True, slots=True)
class UsersUnblockedResult(_UsersUnblockedResult):
    """
    Represents the result of unblocking users operation.

    Attributes:
        removed_users: The users that were successfully unblocked.
    """

    removed_users: tuple[User, ...]


class Result(_Result[_T]):
    """
    This class is used to handle paginated results from the WhatsApp API. You can iterate over the results, and also access the next and previous pages of results.

    - When using the ``next()`` or ``previous()`` methods, the results are returned as a new instance of the :class:`Result` class.
    - You can access the cursors using the ``before`` and ``after`` properties and use them later in the :class:`Pagination` object.

    Example:

        >>> from pywa import WhatsApp, types
        >>> wa = WhatsApp(...)
        >>> res = wa.get_blocked_users(pagination=types.Pagination(limit=100))
        >>> for user in res:
        ...     print(user.name, user.wa_id)
        ...
        >>> if res.has_next:
        ...     next_res = res.next()
        ...
        >>> print(res.all())

    Methods:
        next: Get the next page of results. if there is no next page, it returns empty Result.
        previous: Get the previous page of results. if there is no previous page, it returns empty Result.
        all: Get all results from the current page, previous pages, and next pages.
        empty: Returns an empty Result instance.

    Properties:
        has_next: Check if there is a next page of results.
        has_previous: Check if there is a previous page of results.
        before: Cursor that points to the start of the page of data that has been returned.
        after: Cursor that points to the end of the page of data that has been returned.
    """

    _wa: WhatsAppAsync

    @property
    def empty(self) -> Result[_T]:
        """Returns an empty Result instance."""
        return Result(
            wa=self._wa,
            response={
                "data": [],
                "paging": {"next": self._next_url, "cursors": self._cursors},
            },
            item_factory=self._item_factory,
        )

    async def next(self) -> Result[_T]:
        """
        Get the next page of results. if there is no next page, it returns empty Result.

        - Check if there is a next page using the :attr:`~pywa.types.others.Result.has_next` property before calling this method.
        """
        if self.has_next:
            # noinspection PyProtectedMember
            response = await self._wa.api._make_request(
                method="GET", endpoint=self._next_url
            )
            return Result(
                wa=self._wa, response=response, item_factory=self._item_factory
            )
        return self.empty

    async def previous(self) -> Result[_T]:
        """
        Get the previous page of results. if there is no previous page, it returns empty Result.

        - Check if there is a previous page using the :attr:`~pywa.types.others.Result.has_previous` property before calling this method.
        """
        if self.has_previous:
            # noinspection PyProtectedMember
            response = await self._wa.api._make_request(
                method="GET", endpoint=self._previous_url
            )
            return Result(
                wa=self._wa, response=response, item_factory=self._item_factory
            )
        return self.empty

    async def all(
        self,
        *,
        sleep: float = 0.0,
    ) -> list[_T]:
        """
        Get all results from the current page, previous pages, and next pages.

        - Make sure to provide higher limit in the ``Pagination`` parameter to avoid hitting rate limits.
        - Also consider using the ``sleep`` parameter to avoid hitting rate limits.

        Args:
            sleep: The number of seconds to sleep between requests to avoid hitting rate limits. Default is 0.0 (no sleep).

        Returns:
            A list of all results from the current page, previous pages, and next pages.
        """
        before_data = []
        after_data = []

        prev = self
        while prev.has_previous:
            if sleep > 0:
                await asyncio.sleep(sleep)
            prev = await prev.previous()
            # noinspection PyProtectedMember
            before_data = prev._data + before_data

        next_page = self
        while next_page.has_next:
            if sleep > 0:
                await asyncio.sleep(sleep)
            next_page = await next_page.next()
            after_data += next_page._data

        return before_data + self._data + after_data



from __future__ import annotations

"""
This module contains the errors that can be raised by the WhatsApp Cloud API or incoming error from the webhook.
"""
import warnings
import dataclasses
import functools
from typing import Iterable, ClassVar

import httpx


@dataclasses.dataclass(slots=True, frozen=True)
class WhatsAppError(Exception):
    """
    Base dataclass for WhatsApp errors.

    - `'Cloud API Error Codes' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/support/error-codes>`_.
    - `'Flow Error Codes' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/error-codes>`_.
    - `'Calling Error Codes' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/troubleshooting>`_.
    - `'Block User Error Codes' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/block-users#error-codes>`_.

    Attributes:
        code: The error code.
        message: The error message.
        details: The error details (optional).
        fbtrace_id: The Facebook trace ID (optional).
        href: The href to the documentation (optional).
        raw_response: The :class:`httpx.Response` obj that returned the error (optional, only if the error was raised
            from an API call).
        subcode: The error subcode (optional).
        type: The error type (optional).
        is_transient: Whether the error is transient (optional).
        user_title: The user-facing title for the error (optional).
        user_msg: The user-facing message for the error (optional).
    """

    __error_codes__: ClassVar[Iterable[int] | None] = None

    code: int
    message: str
    details: str | None = None
    fbtrace_id: str | None = None
    href: str | None = None
    raw_response: httpx.Response | None = None
    subcode: int | None = None
    type: str | None = None
    is_transient: bool | None = None
    user_title: str | None = None
    user_msg: str | None = None

    @classmethod
    def from_dict(
        cls, error: dict, response: httpx.Response | None = None
    ) -> WhatsAppError:
        """Create an error from a response."""
        return cls._get_exception(code=(int_code := int(error["code"])))(
            code=int_code,
            message=error["message"],
            details=error.get("error_data", {}).get("details", None),
            fbtrace_id=error.get("fbtrace_id"),
            href=error.get("href"),
            raw_response=response,
            subcode=error.get("error_subcode"),
            type=error.get("type"),
            is_transient=error.get("is_transient"),
            user_title=error.get("error_user_title"),
            user_msg=error.get("error_user_msg"),
        )

    @property
    def status_code(self) -> int | None:
        """The status code (in case of ``raw_response``, else None)."""
        return self.raw_response.status_code if self.raw_response is not None else None

    @classmethod
    def _get_exception(cls, code: int) -> type[WhatsAppError]:
        """Get the exception class from the error code."""
        return _all_exceptions().get(code, WhatsAppError)

    @property
    def error_code(self) -> int:
        """Deprecated, use `code` instead."""
        warnings.warn(
            "WhatsAppError.error_code is deprecated, use WhatsAppError.code instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.code

    @property
    def error_subcode(self) -> int | None:
        """Deprecated, use `subcode` instead."""
        warnings.warn(
            "WhatsAppError.error_subcode is deprecated, use WhatsAppError.subcode instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.subcode

    def __str__(self) -> str:
        """Return a string representation of the error."""
        return self.__repr__()


@functools.cache
def _all_exceptions() -> dict[int, type[WhatsAppError]]:
    """Get all exceptions that can be raised from this error."""
    return {
        code: ssc
        for sc in WhatsAppError.__subclasses__()
        for ssc in sc.__subclasses__()
        for code in ssc.__error_codes__
    }


class AuthorizationError(WhatsAppError):
    """Base exception for all authorization errors."""

    __error_codes__ = None


class AuthException(AuthorizationError):
    """We were unable to authenticate the app user"""

    __error_codes__ = (0,)


class APIMethod(AuthorizationError):
    """Capability or permissions issue."""

    __error_codes__ = (3,)


class PermissionDenied(AuthorizationError):
    """Permission is either not granted or has been removed."""

    __error_codes__ = (10,)


class ExpiredAccessToken(AuthorizationError):
    """Your access token has expired."""

    __error_codes__ = (190,)


class APIPermission(AuthorizationError):
    """Permission is either not granted or has been removed."""

    __error_codes__ = range(200, 300)


# ====================================================================================================


class ThrottlingError(WhatsAppError):
    """Base exception for all rate limit errors."""

    __error_codes__ = None


class ToManyAPICalls(ThrottlingError):
    """The app has reached its API call rate limit."""

    __error_codes__ = (4,)


class RateLimitIssues(ThrottlingError):
    """The WhatsApp Business Account has reached its rate limit."""

    __error_codes__ = (80007,)


class RateLimitHit(ThrottlingError):
    """Cloud API message throughput has been reached."""

    __error_codes__ = (130429,)


class SpamRateLimitHit(ThrottlingError):
    """
    Message failed to send because there are restrictions on how many messages can be sent from this phone number.
    This may be because too many previous messages were blocked or flagged as spam.
    """

    __error_codes__ = (131048,)


class TooManyMessages(ThrottlingError):
    """
    Too many messages sent from the sender phone number to the same recipient phone number in a short period of time.
    """

    __error_codes__ = (131056,)


# ====================================================================================================


class IntegrityError(WhatsAppError):
    """Base exception for all integrity errors."""

    __error_codes__ = None


class TemporarilyBlocked(IntegrityError):
    """
    The WhatsApp Business Account associated with the app has been restricted or disabled for violating a platform policy.
    """

    __error_codes__ = (368,)


class AccountLocked(IntegrityError):
    """
    The WhatsApp Business Account associated with the app has been restricted or disabled for violating a platform
    policy, or we were unable to verify data included in the request against data set on the WhatsApp Business
    Account (e.g, the two-step pin included in the request is incorrect).
    """

    __error_codes__ = (131031,)


class AccountRestrictedFromCountry(IntegrityError):
    """
    The WhatsApp Business Account is restricted from messaging to users in certain countries.

    See WhatsApp Business Messaging Policy for details on allowed countries for messaging in your business category.
    """

    __error_codes__ = (130497,)


# ====================================================================================================


class SendMessageError(WhatsAppError):
    """Base exception for all message errors."""

    __error_codes__ = None


class UserIsInExperimentGroup(SendMessageError):
    """The user is part of an experiment group and the message was not sent as part of an experiment.

    See `Marketing Message Experiment <https://developers.facebook.com/docs/whatsapp/cloud-api/guides/experiments#marketing-message-experiment>`_.
    """

    __error_codes__ = (130472,)


class MessageUndeliverable(SendMessageError):
    """Unable to deliver message. Reasons can include:

    - The recipient phone number is not a WhatsApp phone number.
    - Recipient has not accepted our new Terms of Service and Privacy Policy.
    - Recipient using an old WhatsApp version; must use the following WhatsApp version or greater:
        Android: 2.21.15.15
        SMBA: 2.21.15.15
        iOS: 2.21.170.4
        SMBI: 2.21.170.4
        KaiOS: 2.2130.10
        Web: 2.2132.6
    """

    __error_codes__ = (131026,)


class ReEngagementMessage(SendMessageError):
    """More than 24 hours have passed since the recipient last replied to the sender number."""

    __error_codes__ = (131047,)


class UnsupportedMessageType(SendMessageError):
    """The message type is not supported."""

    __error_codes__ = (131051,)


class MediaDownloadError(SendMessageError):
    """Unable to download the media sent by the user."""

    __error_codes__ = (131052,)


class MediaUploadError(SendMessageError):
    """Unable to upload the media used in the message."""

    __error_codes__ = (131053,)


class RecipientNotInAllowedList(SendMessageError):
    """
    When using test numbers, you can add up to 5 numbers that can receive messages. This error is raised when the
    recipient is not in the list.
    """

    __error_codes__ = (131030,)


class InvalidParameter(SendMessageError):
    """The parameter you passed is invalid."""

    __error_codes__ = (131009,)


class MissingRequiredParameter(SendMessageError):
    """You must provide a value for the required parameter."""

    __error_codes__ = (131008,)


class TemplateParamCountMismatch(SendMessageError):
    """The number of variable parameter values included in the request did not match the number of variable parameters
    defined in the template.."""

    __error_codes__ = (132000,)


class TemplateNotExists(SendMessageError):
    """The template does not exist in the specified language or the template has not been approved."""

    __error_codes__ = (132001,)


class TemplateTextTooLong(SendMessageError):
    """The template text is too long."""

    __error_codes__ = (132005,)


class TemplateContentPolicyViolation(SendMessageError):
    """The template content violates a WhatsApp policy."""

    __error_codes__ = (132007,)


class TemplateParamValueInvalid(SendMessageError):
    """Variable parameter values formatted incorrectly."""

    __error_codes__ = (132008,)


class TemplateParamFormatMismatch(SendMessageError):
    """Variable parameter values formatted incorrectly."""

    __error_codes__ = (132012,)


class TemplatePaused(SendMessageError):
    """Template is paused due to low quality so it cannot be sent in a template message."""

    __error_codes__ = (132015,)


class TemplateDisabled(SendMessageError):
    """Template has been paused too many times due to low quality and is now permanently disabled."""

    __error_codes__ = (132016,)


class FlowBlocked(SendMessageError):
    """Flow is in blocked state."""

    __error_codes__ = (132068,)


class FlowThrottled(SendMessageError):
    """Flow is in throttled state and 10 messages using this flow were already sent in the last hour."""

    __error_codes__ = (132069,)


class GenericError(SendMessageError):
    """Generic error."""

    __error_codes__ = (135000,)


class UnknownError(SendMessageError):
    """Message failed to send due to an unknown error."""

    __error_codes__ = (131000,)


class AccessDenied(SendMessageError):
    """Permission is either not granted or has been removed."""

    __error_codes__ = (131005,)


class ServiceUnavailable(SendMessageError):
    """A service is temporarily unavailable."""

    __error_codes__ = (131016,)


class RecipientCannotBeSender(SendMessageError):
    """Sender and recipient phone number is the same."""

    __error_codes__ = (131021,)


class BusinessPaymentIssue(SendMessageError):
    """Message failed to send because there were one or more errors related to your payment method."""

    __error_codes__ = (131042, 131044)  # 131044 is for calling


class IncorrectCertificate(SendMessageError):
    """Message failed to send due to a phone number registration error."""

    __error_codes__ = (131045,)


class AccountInMaintenanceMode(SendMessageError):
    """Business Account is in maintenance mode"""

    __error_codes__ = (131057,)


class UserStoppedMarketingMessages(SendMessageError):
    """User has stopped marketing messages from the business account. you should listen to UserMarketingPreferences updates"""

    __error_codes__ = (131050,)


# ====================================================================================================


class FlowError(WhatsAppError):
    """
    Base exception for all flow errors.

     Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/flows/reference/error-codes>`_.
    """

    __error_codes__ = None


class FlowBlockedByIntegrity(FlowError):
    """Unfortunately, we've identified an issue with integrity in your account and have prevented you from creating
    or publishing your Flow."""

    __error_codes__ = (139000,)


class FlowUpdatingError(FlowError):
    """
    Flow failed to update

    - Flow can't be updated (You attempted to update a Flow that has already been published.)
    - Error while processing Flow JSON.
    - Specify Endpoint Uri in Flow JSON (You provided endpoint_uri param for a Flow with Flow JSON version below 3.0.)
    """

    __error_codes__ = (139001,)


class FlowPublishingError(FlowError):
    """
    Flow failed to publish.

    - Publishing Flow in invalid state
    - Publishing Flow with validation errors
    - Publishing Flow without data_channel_uri
    - Publishing without specifying endpoint_uri is forbidden
      (Starting from Flow JSON version 3.0 endpoint_uri should be specified via API.)
    - Versions in Flow JSON file are not available for publishing
    - No Phone Number connected to WhatsApp Business Account
    - Missing Flows Signed Public Key
    - No Application Connected to the Flow
    - Endpoint Not Available
    - WhatsApp Business Account is not subscribed to Flows Webhooks
    """

    __error_codes__ = (139002,)


class FlowDeprecatingError(FlowError):
    """
    Flow failed to deprecate.

    - Can't deprecate unpublished flow
    - Flow is already deprecated
    """

    __error_codes__ = (139003,)


class FlowDeletingError(FlowError):
    """
    Flow failed to delete.

    - Can't delete published Flow
    """

    __error_codes__ = (139004,)


# ====================================================================================================


class BlockUserError(WhatsAppError):
    """
    Base exception for all block user errors.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/block-users#error-codes>`_.
    """

    __error_codes__ = None


class BulkBlockingFailed(BlockUserError):
    """
    Failed to block/unblock some users

    Bulk blocking failed to block some or all of the users.
    """

    __error_codes__ = (139100,)


class BlockListLimitReached(BlockUserError):
    """
    Blocklist limit reached

    The blocklist limit is reached when the 64k limit is met.
    """

    __error_codes__ = (139101,)


class BlockListConcurrentUpdate(BlockUserError):
    """
    Blocklist concurrent update

    Occurs when the block list is updated while performing a pagination request and version_id does not match.
    """

    __error_codes__ = (139102,)


class BlockUserInternalError(BlockUserError):
    """
    Internal error

    Internal error, please try again.
    """

    __error_codes__ = (139103,)


# ====================================================================================================


class CallingError(WhatsAppError):
    """
    Base exception for all calling errors.

    - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/troubleshooting>`_.
    """

    __error_codes__ = None


class CallingNotEnabled(CallingError):
    """
    Calling is not enabled on the business phone number.

    `Configure call settings to enable Calling API features <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/call-settings>`_.
    """

    __error_codes__ = (138000,)


class CallingCannotBeEnabled(CallingError):
    """
    WhatsApp Business calling cannot be enabled because technical pre-requisites are not met.

    See `prerequisites <https://developers.facebook.com/docs/whatsapp/cloud-api/calling#step-1--prerequisites>`_ for more details

    `Configure SIP <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/sip>`_ or ensure there is at-least one app subscribed to the WhatsApp Business Account that also has subscription to the calls webhook field.
    """

    __error_codes__ = (138018,)


class ReceiverUncallable(CallingError):
    """
    Receiver is unable to receive calls

    Reasons can include:

    - The recipient phone number is not a WhatsApp phone number.
    - The recipient has not accepted our new Terms of Service and Privacy Policy.
    - Recipient using an unsupported client. The currently supported clients are Android, iOS, SMB Android and SMB iOS

    Confirm with the recipient that they agree to be contacted by you over WhatsApp and are using the latest version of WhatsApp.
    """

    __error_codes__ = (138001,)


class ConcurrentCallsLimit(CallingError):
    """
    Limit reached for maximum concurrent calls (1000) for the given number

    Try again later or reduce the frequency or amount of API calls the app is making.
    """

    __error_codes__ = (138002,)


class DuplicateCall(CallingError):
    """
    A call is already ongoing with the receiver

    Try again later when the current call ends.
    """

    __error_codes__ = (138003,)


class CallConnectionError(CallingError):
    """
    Error while connecting the call

    Try again later or investigate the connection params provided to the API.
    """

    __error_codes__ = (138004,)


class CallRateLimitExceeded(CallingError):
    """
    Limit reached for maximum calls that can be initiated by the business phone number

    Try again later or reduce the frequency or amount of API calls the app is making.
    """

    __error_codes__ = (138005,)


class CallPermissionNotFound(CallingError):
    """
    No approved `call permission <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-call-permissions>`_ from the recipient

    Ensure a call permission has been accepted by the consumer
    """

    __error_codes__ = (138006,)


class CallConnectionTimeout(CallingError):
    """
    Call was unable to connect due to a timeout

    Business did not apply the offer/answer SDP from Cloud API in time. Connect API was not invoked with the answer SDP in time
    """

    __error_codes__ = (138007,)


class CallPermissionRequestLimitHit(CallingError):
    """
    Limit reached for call permission request sends for the given business and consumer pair

    When a business sends more than the limit of call permission requests per time period, Call Permission Requests are rate limited. A connected call with a consumer will reset the limits.
    """

    __error_codes__ = (138009,)


class BusinessInitiatedCallsLimitHit(CallingError):
    """
    Limit reached for maximum business initiated calls allowed in 24 hours. Currently 5 connected business initiated calls are allowed within 24 hours.

    Exact error details will be listed in the error_data section of the response payload. Details will include a timestamp when the next call is allowed.
    """

    __error_codes__ = (138012,)


class FetchCallPermissionLimitHit(CallingError):
    """
    Limit reached for requests to the fetch call permission status API.

    Try again later or reduce the frequency or amount of requests to the API the app is making.
    """

    __error_codes__ = (
        138013,
        613,
    )  # WhatsApp changed the error code from 138013 to 613



from __future__ import annotations

from . import MessageType

"""This module contains types related to WhatsApp calls, including call connection, termination, and status updates."""

__all__ = [
    "CallPermissionUpdate",
    "CallPermissionResponse",
    "CallPermissionResponseSource",
    "CallConnect",
    "SessionDescription",
    "CallEvent",
    "CallDirection",
    "CallTerminate",
    "CallTerminateStatus",
    "CallStatus",
    "CallStatusType",
    "CallingSettingsStatus",
    "CallIconVisibility",
    "CallbackPermissionStatus",
    "SIPStatus",
    "SIPServer",
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday",
    "HolidaySchedule",
    "CallHours",
    "CallingSettings",
    "BusinessPhoneNumberSettings",
    "CallPermissionStatus",
    "CallPermissionActionLimit",
    "CallPermissionAction",
    "CallPermission",
    "CallPermissionActionLimit",
    "CallPermissionAction",
    "CallPermissions",
]

import dataclasses
import datetime
from typing import TYPE_CHECKING, Generic

from .base_update import BaseUserUpdate, BaseUpdate, _ClientShortcuts  # noqa
from .others import (
    Metadata,
    SuccessResult,
    ReplyToMessage,
    StorageConfiguration,
)
from .callback import _CallbackDataT, CallbackData
from .. import utils
from ..errors import WhatsAppError

if TYPE_CHECKING:
    from ..client import WhatsApp
    from .sent_update import InitiatedCall


class _CallShortcuts:
    """Base class for call actions."""

    id: str
    _client: WhatsApp
    _internal_sender: str
    _internal_recipient: str

    @property
    def message_id_to_reply(self) -> str:
        """Raises an error because call terminate updates cannot be replied."""
        raise ValueError("You cannot use `message_id_to_reply` to quote a call update.")

    @property
    def caller(self) -> str:
        """
        The WhatsApp ID of the business phone number that initiated the call.
        """
        return self._internal_recipient

    @property
    def callee(self) -> str:
        """The WhatsApp ID of the user that received the call."""
        return self._internal_sender

    def pre_accept(self, *, sdp: SessionDescription) -> SuccessResult:
        """
        Pre-accept the call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#pre-accept-call>`_.

        In essence, when you pre-accept an inbound call, you are allowing the calling media connection to be established before attempting to send call media through the connection.

        When you then call the accept call endpoint, media begins flowing immediately since the connection has already been established

        Pre-accepting calls is recommended because it facilitates faster connection times and avoids `audio clipping issues <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/troubleshooting#audio-clipping-issue-and-solution>`_.

        There is about 30 to 60 seconds after the Call Connect webhook is sent for the business to accept the phone call. If the business does not respond, the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.

        Returns:
            Whether the call was pre-accepted.
        """
        return self._client.pre_accept_call(call_id=self.id, sdp=sdp)

    def accept(
        self,
        sdp: SessionDescription,
        *,
        tracker: str | CallbackData | None = None,
    ) -> SuccessResult:
        """
        Connect to a call by providing a call agent's SDP.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#accept-call>`_.

        You have about 30 to 60 seconds after the Call Connect Webhook is sent to accept the phone call. If your business does not respond, the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.
            tracker: The data to track the call with (optional, up to 512 characters, for complex data You can use :class:`CallbackData`).

        Returns:
            Whether the call was accepted.
        """
        return self._client.accept_call(call_id=self.id, sdp=sdp, tracker=tracker)

    def reject(self) -> SuccessResult:
        """
        Reject the call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#reject-call>`_.

        You have about 30 to 60 seconds after the Call Connect webhook is sent to accept the phone call. If the business does not respond the call is terminated on the WhatsApp user side with a "Not Answered" notification and a Terminate Webhook is delivered back to you.

        Returns:
            Whether the call was rejected.
        """
        return self._client.reject_call(call_id=self.id)

    def terminate(self) -> SuccessResult:
        """
        Terminate the active call.

        - Read more at `developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-initiated-calls#terminate-call>`_.

        This must be done even if there is an RTCP BYE packet in the media path. Ending the call this way also ensures pricing is more accurate.
        When the WhatsApp user terminates the call, you do not have to call this endpoint. Once the call is successfully terminated, a Call Terminate Webhook will be sent to you.

        Returns:
            Whether the call was terminated.
        """
        return self._client.terminate_call(call_id=self.id)


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallConnect(BaseUserUpdate, _CallShortcuts):
    """
    Represents a call connection event.

    - This update arrives when a call is initiated by the business or the user.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        from_user: The user who participated in the call, either as caller or callee.
        timestamp: The timestamp when this call was made (in UTC).
        event: The calling event (always ``CONNECT``).
        direction: Whether the call was initiated by the business or the user.
        session: The session information, including SDP type and SDP info.
        shared_data: Shared data between handlers.
    """

    event: CallEvent
    direction: CallDirection
    session: SessionDescription | None

    _webhook_field = "calls"

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> CallConnect:
        call = (value := (entry := update["entry"][0])["changes"][0]["value"])["calls"][
            0
        ]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=call["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            from_user=client._usr_cls.from_dict(value["contacts"][0], client=client)
            if "contacts" in value  # Only available for USER_INITIATED calls
            else client._usr_cls(_client=client, wa_id=call["to"], name=None),
            timestamp=datetime.datetime.fromtimestamp(
                int(call["timestamp"]),
                datetime.timezone.utc,
            ),
            event=CallEvent(call["event"]),
            direction=CallDirection(call["direction"]),
            session=SessionDescription.from_dict(call["session"])
            if "session" in call
            else None,
        )


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallPermissionUpdate(BaseUserUpdate):
    """
    Represents a call permission update.

    - This update arrives when a call permission request is sent to the user, and the user responds with an action.

    Attributes:
        id: The message ID.
        metadata: The metadata of the message (to which phone number this call permission request was sent).
        type: The type of the message (always ``INTERACTIVE``).
        reply_to_message: The message that this call permission request is replying to.
        from_user: The user who acted on the call permission request.
        timestamp: The timestamp when this update is sent (in UTC).
        response: The response to the call permission request (either ``ACCEPT`` or ``REJECT``).
        response_source: The source of the call permission response (either ``USER_ACTION`` or ``AUTOMATIC``).
        expiration_timestamp: The timestamp when the call permission request expires (if applicable).
    """

    type: MessageType
    reply_to_message: ReplyToMessage
    response: CallPermissionResponse
    response_source: CallPermissionResponseSource
    expiration_timestamp: datetime.datetime | None = None

    _webhook_field = "messages"

    def __bool__(self):
        """
        Returns True if the call permission request was accepted.
        """
        return self.response != CallPermissionResponse.REJECT

    @property
    def is_expired(self) -> bool:
        """
        Whether the call permission request has expired.
        """
        return (
            self.expiration_timestamp is not None
            and self.expiration_timestamp < datetime.datetime.now(datetime.timezone.utc)
        )

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> CallPermissionUpdate:
        perm = (
            msg := (value := (entry := update["entry"][0])["changes"][0]["value"])[
                "messages"
            ][0]
        )["interactive"]["call_permission_reply"]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=msg["id"],
            type=MessageType(msg["type"]),
            metadata=Metadata.from_dict(value["metadata"]),
            from_user=client._usr_cls(
                _client=client,
                wa_id=msg["from"],
                name=None,
            ),
            timestamp=datetime.datetime.fromtimestamp(
                int(msg["timestamp"]),
                datetime.timezone.utc,
            ),
            reply_to_message=ReplyToMessage.from_dict(msg["context"])
            if msg.get("context", {}).get("id")
            else None,
            response=CallPermissionResponse(perm["response"]),
            response_source=CallPermissionResponseSource(perm["response_source"]),
            expiration_timestamp=datetime.datetime.fromtimestamp(
                int(perm.get("expiration_timestamp", 0)),
                datetime.timezone.utc,
            )
            if "expiration_timestamp" in perm
            else None,
        )


class CallPermissionResponse(utils.StrEnum):
    """
    Represents the response to a call permission request.

    Attributes:
        ACCEPT: The user accepted the call permission request.
        REJECT: The user rejected the call permission request.
    """

    _check_value = str.islower
    _modify_value = str.lower

    ACCEPT = "accept"
    REJECT = "reject"

    UNKNOWN = "UNKNOWN"


class CallPermissionResponseSource(utils.StrEnum):
    """
    Represents the source of the call permission response.

    Attributes:
        USER_ACTION: User approved or rejected the permission
        AUTOMATIC: An automatic permission approval due to the WhatsApp user initiating the call
    """

    _check_value = str.islower
    _modify_value = str.lower

    USER_ACTION = "user_action"
    AUTOMATIC = "automatic"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class SessionDescription(utils.FromDict):
    """
    Contains the Session Description Protocol (SDP) type and description language.

    - `Learn more about Session Description Protocol (SDP) <https://www.rfc-editor.org/rfc/rfc8866.html?fbclid=IwZXh0bgNhZW0CMTAAYnJpZBExZ29GWHBsSlEyRXA1aThYMgEeF-jEVWkTrJf0tEEbpY0iS9kkCIFncddV7i-RU-qm_eAD0MG6oE3oIwsNUWY_aem_X06J-leL4yOl1-TWuaMyoQ>`_.
    - `View example SDP structures <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/reference#sdp-overview-and-sample-sdp-structures>`_.

    Attributes:
        sdp_type: "offer", to indicate SDP offer
        sdp: The SDP info of the device on the other end of the call. The SDP must be compliant with `RFC 8866 <https://datatracker.ietf.org/doc/html/rfc8866?fbclid=IwZXh0bgNhZW0CMTAAYnJpZBExaENMd1p1NEpTbkQyQlNpMgEeyPkNeWi6AvgR4q8vdCIMx6rJ4AWd4yxRYpn6mvZ31VjC_0fN6FtPdrxtDOs_aem_7po_7yOAFkrlcV4Xn3TVnQ>`_.
    """

    sdp_type: str
    sdp: str

    def to_dict(self) -> dict[str, str]:
        return {
            "sdp_type": self.sdp_type,
            "sdp": self.sdp,
        }


class CallEvent(utils.StrEnum):
    """
    Represents the type of call event.

    Attributes:
        CONNECT: The call is connected.
        TERMINATE: The call is terminated.
    """

    _check_value = str.islower
    _modify_value = str.lower

    CONNECT = "connect"
    TERMINATE = "terminate"

    UNKNOWN = "UNKNOWN"


class CallDirection(utils.StrEnum):
    """
    Represents the direction of a call (who initiated the call).

    Attributes:
        BUSINESS_INITIATED: The call was initiated by the business.
        USER_INITIATED: The call was initiated by the user.
    """

    BUSINESS_INITIATED = "BUSINESS_INITIATED"
    USER_INITIATED = "USER_INITIATED"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallTerminate(BaseUserUpdate, _CallShortcuts, Generic[_CallbackDataT]):
    """
    Represents a call termination event.

    - This update arrives when a call is terminated, either by the business or the user.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        from_user: The user who participated in the call, either as caller or callee.
        timestamp: The timestamp when this update is sent (in UTC).
        event: The calling event (always ``TERMINATE``).
        direction: Whether the call was initiated by the business or the user.
        status: The status of the call (either ``FAILED`` or ``COMPLETED``).
        start_time: The start time of the call in UTC (Only if the call was picked up).
        end_time: The end time of the call in UTC (Only if the call was picked up).
        duration: The duration of the call in seconds (Only if the call was picked up).
        tracker: The tracker that the call is initiated with.
        shared_data: Shared data between handlers.
    """

    event: CallEvent
    direction: CallDirection
    status: CallTerminateStatus
    start_time: datetime.datetime | None
    end_time: datetime.datetime | None
    duration: int | None
    error: WhatsAppError | None
    tracker: _CallbackDataT | None

    _webhook_field = "calls"

    @property
    def message_id_to_reply(self) -> str:
        """Raises an error because call terminate updates cannot be replied."""
        raise ValueError(
            "You cannot use `message_id_to_reply` to quote a call terminate update."
        )

    @classmethod
    def from_update(
        cls, client: WhatsApp, update: dict
    ) -> CallTerminate[_CallbackDataT]:
        call = (value := (entry := update["entry"][0])["changes"][0]["value"])["calls"][
            0
        ]
        error = value.get("errors", (None,))[0]
        direction = CallDirection(call["direction"])
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=call["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            from_user=client._usr_cls(
                _client=client,
                wa_id=call["from"]
                if direction == CallDirection.BUSINESS_INITIATED
                else call["to"],
                name=None,
            ),
            timestamp=datetime.datetime.fromtimestamp(
                int(call["timestamp"]),
                datetime.timezone.utc,
            ),
            event=CallEvent(call["event"]),
            direction=direction,
            status=CallTerminateStatus(call["status"]),
            start_time=datetime.datetime.fromtimestamp(
                int(call["start_time"]), datetime.timezone.utc
            )
            if "start_time" in call
            else None,
            end_time=datetime.datetime.fromtimestamp(
                int(call["end_time"]), datetime.timezone.utc
            )
            if "end_time" in call
            else None,
            duration=int(call["duration"]) if "duration" in call else None,
            error=WhatsAppError.from_dict(error) if error else None,
            tracker=call.get("biz_opaque_callback_data"),
        )

    def recall(
        self, sdp: SessionDescription, *, tracker: str | CallbackData | None = None
    ) -> InitiatedCall:
        """
        Recall the call with the given SDP.

        - This is useful if you want to re-initiate a call after it has been terminated.

        Args:
            sdp: Contains the session description protocol (SDP) type and description language.
            tracker: The data to track the call with (optional, up to 512 characters, for complex data You can use :class:`CallbackData`).

        Returns:
            An InitiatedCall object containing the details of the initiated call.
        """
        return self._client.initiate_call(
            to=self._internal_sender,
            sdp=sdp,
            tracker=tracker,
            phone_id=self._internal_recipient,
        )


class CallTerminateStatus(utils.StrEnum):
    """
    Represents the status of a call termination event.

    Attributes:
        FAILED: The call failed.
        COMPLETED: The call was completed successfully.
    """

    FAILED = "FAILED"
    COMPLETED = "COMPLETED"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallStatus(BaseUserUpdate, _CallShortcuts, Generic[_CallbackDataT]):
    """
    Represents a call status update.

    - This update arrives when during a business-initiated call, the user either accepts or rejects the call.

    Attributes:
        id: The call ID.
        metadata: The metadata of the call (to which phone number this call was made or received).
        timestamp: The timestamp when this status update is sent (in UTC).
        type: The type of the status update (always "call").
        status: The status of the call (either "RINGING", "ACCEPTED", or "REJECTED").
        tracker: The tracker that the call is initiated with.
        shared_data: Shared data between handlers.
    """

    type: str
    status: CallStatusType
    tracker: _CallbackDataT | None = None

    _webhook_field = "calls"

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> CallStatus:
        status = (value := (entry := update["entry"][0])["changes"][0]["value"])[
            "statuses"
        ][0]
        return cls(
            _client=client,
            raw=update,
            waba_id=entry["id"],
            id=status["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            timestamp=datetime.datetime.fromtimestamp(
                int(status["timestamp"]),
                datetime.timezone.utc,
            ),
            from_user=client._usr_cls(
                wa_id=status["recipient_id"], name=None, _client=client
            ),
            type=status["type"],
            status=CallStatusType(status["status"]),
            tracker=status.get("biz_opaque_callback_data"),
        )


class CallStatusType(utils.StrEnum):
    """
    Represents the type of call status.

    Attributes:
        RINGING: Business initiated call is ringing the user.
        ACCEPTED: Business initiated call is accepted by the user.
        REJECTED: Business initiated call is rejected by the user.
    """

    RINGING = "RINGING"
    ACCEPTED = "ACCEPTED"
    REJECTED = "REJECTED"

    UNKNOWN = "UNKNOWN"


class CallingSettingsStatus(utils.StrEnum):
    """
    Represents the status of calling settings.
    """

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"

    UNKNOWN = "UNKNOWN"


class CallIconVisibility(utils.StrEnum):
    """
    Represents the visibility of the call icon.
    """

    DEFAULT = "DEFAULT"
    DISABLE_ALL = "DISABLE_ALL"

    UNKNOWN = "UNKNOWN"


class CallbackPermissionStatus(utils.StrEnum):
    """
    Represents the status of callback permission.
    """

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"

    UNKNOWN = "UNKNOWN"


class SIPStatus(utils.StrEnum):
    """
    Represents the status of SIP (Session Initiation Protocol).
    """

    ENABLED = "ENABLED"
    DISABLED = "DISABLED"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class SIPServer(utils.FromDict):
    """
    Represents a SIP server configuration.

    Attributes:
        hostname: The hostname of the SIP server.
        port: The port of the SIP server.
        request_uri_user_params: Optional parameters for the request URI user.
        sip_user_password: The password for the SIP user (only if ``include_sip_credentials`` is True).
    """

    hostname: str
    port: int
    request_uri_user_params: dict[str, str] | None
    sip_user_password: str | None


@dataclasses.dataclass(slots=True, kw_only=True)
class WeekDay:
    """
    Represents a day of the week with its opening and closing times.

    - You can use the subclasses for specific days (e.g., :class:`Monday`, ``Tuesday``, etc.) to create instances with predefined day names.

    Attributes:
        day_of_week: The day of the week (e.g., "MONDAY", "TUESDAY").
        open_time: The opening time in 24-hour format (e.g., "0400" for 4:00 AM).
        close_time: The closing time in 24-hour format (e.g., "2200" for 10:00 PM).
    """

    day_of_week: str
    open_time: datetime.time
    close_time: datetime.time

    def to_dict(self) -> dict[str, str]:
        return {
            "day_of_week": self.day_of_week,
            "open_time": self.open_time.strftime("%H%M"),
            "close_time": self.close_time.strftime("%H%M"),
        }


@dataclasses.dataclass(slots=True, kw_only=True)
class Monday(WeekDay):
    """
    Represents Monday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="MONDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Tuesday(WeekDay):
    """
    Represents Tuesday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="TUESDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Wednesday(WeekDay):
    """
    Represents Wednesday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="WEDNESDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Thursday(WeekDay):
    """
    Represents Thursday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="THURSDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Friday(WeekDay):
    """
    Represents Friday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="FRIDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Saturday(WeekDay):
    """
    Represents Saturday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="SATURDAY", init=False, repr=False)


@dataclasses.dataclass(slots=True, kw_only=True)
class Sunday(WeekDay):
    """
    Represents Sunday with its opening and closing times.
    """

    day_of_week: str = dataclasses.field(default="SUNDAY", init=False, repr=False)


_day_to_weekday_map = {
    "MONDAY": Monday,
    "TUESDAY": Tuesday,
    "WEDNESDAY": Wednesday,
    "THURSDAY": Thursday,
    "FRIDAY": Friday,
    "SATURDAY": Saturday,
    "SUNDAY": Sunday,
}


def _day_to_weekday(day: dict[str, str]) -> WeekDay:
    """
    Converts a dictionary representation of a day to a WeekDay object.

    Args:
        day: A dictionary containing the day of the week, open time, and close time.

    Returns:
        An instance of a specific WeekDay subclass based on the day of the week.
    """
    # noinspection PyArgumentList
    return _day_to_weekday_map[day["day_of_week"]](
        open_time=datetime.datetime.strptime(day["open_time"], "%H%M").time(),
        close_time=datetime.datetime.strptime(day["close_time"], "%H%M").time(),
    )


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class HolidaySchedule:
    """
    Represents a holiday schedule with a date and opening/closing times.

    Attributes:
        date: The date of the holiday.
        start_time: The opening time in 24-hour format (e.g., "0000" for midnight).
        end_time: The closing time in 24-hour format (e.g., "2359" for 11:59 PM).
    """

    date: datetime.date
    start_time: datetime.time
    end_time: datetime.time

    def to_dict(self) -> dict[str, str]:
        return {
            "date": self.date.strftime("%Y-%m-%d"),
            "start_time": self.start_time.strftime("%H%M"),
            "end_time": self.end_time.strftime("%H%M"),
        }

    @classmethod
    def from_dict(cls, data: dict) -> HolidaySchedule:
        return cls(
            date=datetime.datetime.strptime(data["date"], "%Y-%m-%d").date(),
            start_time=datetime.datetime.strptime(data["start_time"], "%H%M").time(),
            end_time=datetime.datetime.strptime(data["end_time"], "%H%M").time(),
        )


@dataclasses.dataclass(slots=True, kw_only=True, frozen=True)
class CallHours:
    """
    Represents the call hours settings for a business.

    - Maximum of 2 entries allowed per day of week
    - open_time must be before close_time
    - Overlapping entries not allowed

    Attributes:
        status: Enable or disable the call hours for the business (If call hours are disabled, the business is considered open all 24 hours of the day, 7 days a week).
        timezone_id: The timezone that the business is operating within. `Learn more about supported values for timezone_id <https://developers.facebook.com/docs/facebook-business-extension/fbe/reference#time-zones>`_.
        weekly_operating_hours: The operating hours schedule for each day of the week.
        holiday_schedule: An optional override to the weekly schedule, Up to 20 overrides can be specified. Note: If holiday_schedule is not passed in the request, then the existing holiday_schedule will be deleted and replaced with an empty schedule.
    """

    status: CallingSettingsStatus
    timezone_id: str
    weekly_operating_hours: list[WeekDay]
    holiday_schedule: list[HolidaySchedule] | None = None

    def to_dict(self) -> dict:
        return {
            "status": self.status.value,
            "timezone_id": self.timezone_id,
            "weekly_operating_hours": [
                day.to_dict() for day in self.weekly_operating_hours
            ],
            "holiday_schedule": [holiday.to_dict() for holiday in self.holiday_schedule]
            if self.holiday_schedule
            else None,
        }

    @classmethod
    def from_dict(cls, data: dict) -> CallHours:
        return cls(
            status=CallingSettingsStatus(data["status"]),
            timezone_id=data["timezone_id"],
            weekly_operating_hours=[
                _day_to_weekday(day) for day in data["weekly_operating_hours"]
            ],
            holiday_schedule=[
                HolidaySchedule.from_dict(holiday)
                for holiday in data.get("holiday_schedule", [])
            ]
            if "holiday_schedule" in data
            else None,
        )


class STRPKeyExchangeProtocol(utils.StrEnum):
    """
    Represents the SRTP key exchange protocol.

    Attributes:
        DTLS: Datagram Transport Layer Security (DTLS) is used for key exchange.
        SDES: Session Description Protocol Security Descriptions (SDES) is used for key exchange.
    """

    DTLS = "DTLS"
    SDES = "SDES"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(slots=True, kw_only=True)
class CallingSettings:
    """
    Represents the calling settings for a business phone number.

    - See `Configure Call Settings <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/call-settings>`_ for more details.

    Attributes:
        status: Enable or disable the Calling API for the given business phone number.
        call_icon_visibility: Configure whether the WhatsApp call button icon displays for users when chatting with the business. See `Call Icon Visibility <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/call-settings#parameter-details>`_ for more details.
        call_hours: Allows you specify and trigger call settings for incoming calls based on your timezone, business operating hours, and holiday schedules.
        callback_permission_status: Configure whether a WhatsApp user is prompted with a call permission request after calling your business.
        sip: Configure call signaling via signal initiation protocol (SIP). Note: When SIP is enabled, you cannot use calling related endpoints and will not receive calling related webhooks.
    """

    status: CallingSettingsStatus | None = None
    call_icon_visibility: CallIconVisibility | None = None
    call_hours: CallHours | None = None
    callback_permission_status: CallbackPermissionStatus | None = None
    srtp_key_exchange_protocol: STRPKeyExchangeProtocol | None = None
    sip: SIPServer | None = None

    def to_dict(self):
        data = {}
        if self.status:
            data["status"] = self.status.value
        if self.call_icon_visibility:
            data["call_icon_visibility"] = self.call_icon_visibility.value
        if self.callback_permission_status:
            data["callback_permission_status"] = self.callback_permission_status.value
        if self.call_hours:
            data["call_hours"] = self.call_hours.to_dict()
        if self.srtp_key_exchange_protocol:
            data["srtp_key_exchange_protocol"] = self.srtp_key_exchange_protocol.value
        if self.sip:
            data["sip"] = dataclasses.asdict(self.sip)
        return data

    @classmethod
    def from_dict(cls, data: dict) -> CallingSettings:
        return cls(
            status=CallingSettingsStatus(data["status"]) if "status" in data else None,
            call_icon_visibility=CallIconVisibility(data["call_icon_visibility"])
            if "call_icon_visibility" in data
            else None,
            call_hours=CallHours.from_dict(data["call_hours"])
            if "call_hours" in data
            else None,
            callback_permission_status=CallbackPermissionStatus(
                data["callback_permission_status"]
            )
            if "callback_permission_status" in data
            else None,
            srtp_key_exchange_protocol=STRPKeyExchangeProtocol(
                data["srtp_key_exchange_protocol"]
            )
            if "srtp_key_exchange_protocol" in data
            else None,
            sip=SIPServer.from_dict(data["sip"]) if "sip" in data else None,
        )


@dataclasses.dataclass(slots=True, kw_only=True)
class BusinessPhoneNumberSettings(utils.APIObject):
    """
    Represents the settings of a WhatsApp Business Phone Number.

    Attributes:
        calling: The calling settings of the phone number.
        storage_configuration: The storage configuration for the phone number.
    """

    calling: CallingSettings | None = None
    storage_configuration: StorageConfiguration

    @classmethod
    def from_dict(cls, data: dict) -> BusinessPhoneNumberSettings:
        return cls(
            calling=CallingSettings.from_dict(data.get("calling"))
            if data.get("calling")
            else None,
            storage_configuration=StorageConfiguration.from_dict(
                data["storage_configuration"]
            ),
        )

    def to_dict(self) -> dict:
        return {"calling": self.calling.to_dict() if self.calling else None}


class CallPermissionStatus(utils.StrEnum):
    """
    Represents the status of a call permission.

    Attributes:
        NO_PERMISSION: No permission granted.
        TEMPORARY: Temporary permission granted.
    """

    _check_value = str.islower
    _modify_value = str.lower

    NO_PERMISSION = "no_permission"
    TEMPORARY = "temporary"

    UNKNOWN = "UNKNOWN"


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallPermissionActionLimit:
    """
    Represents a limit for a specific action that can be performed with the call permission.

    Attributes:
        time_period: The time period for which the limit applies, represented in ISO 8601 format (e.g., "PT24H" for 24 hours).
        max_allowed: The maximum number of actions allowed within the specified time period.
        current_usage: The current number of actions taken within the specified time period.
        limit_expiration_time: The Unix time at which the limit will expire in UTC timezone (optional).
    """

    time_period: str
    max_allowed: int
    current_usage: int
    limit_expiration_time: datetime.datetime | None = None

    @classmethod
    def from_dict(cls, data: dict) -> CallPermissionActionLimit:
        return cls(
            time_period=data["time_period"],
            max_allowed=data["max_allowed"],
            current_usage=data["current_usage"],
            limit_expiration_time=datetime.datetime.fromtimestamp(
                data["limit_expiration_time"], datetime.timezone.utc
            )
            if "limit_expiration_time" in data
            else None,
        )


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallPermissionAction:
    """
    Represents an action that can be performed with the call permission.

    Attributes:
        action_name: The name of the action (e.g., "send_call_permission_request").
        can_perform_action: Whether the action can be performed.
        limits: A list of limits associated with the action.
    """

    action_name: str
    can_perform_action: bool
    limits: list[CallPermissionActionLimit]

    @classmethod
    def from_dict(cls, data: dict) -> CallPermissionAction:
        return cls(
            action_name=data["action_name"],
            can_perform_action=data["can_perform_action"],
            limits=[
                CallPermissionActionLimit.from_dict(limit) for limit in data["limits"]
            ],
        )


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallPermission:
    """
    Represents a call permission with its status and expiration time.

    Attributes:
        status: The current status of the permission.
        expiration_time: The time at which the permission will expire in UTC timezone.
    """

    status: CallPermissionStatus
    expiration_time: datetime.datetime | None = None

    @classmethod
    def from_dict(cls, data: dict) -> CallPermission:
        return cls(
            status=CallPermissionStatus(data["status"]),
            expiration_time=datetime.datetime.fromtimestamp(
                data["expiration_time"], datetime.timezone.utc
            )
            if "expiration_time" in data
            else None,
        )


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallPermissions:
    """
    Represents the call permissions to specific WhatsApp user.

    - Read more about `Call Permissions <https://developers.facebook.com/docs/whatsapp/cloud-api/calling/user-call-permissions>`_.

    Attributes:
        permission: The call permission status and expiration time.
        actions: A list of actions that can be performed with the call permission.
    """

    permission: CallPermission
    actions: list[CallPermissionAction]

    @classmethod
    def from_dict(cls, data: dict) -> CallPermissions:
        return cls(
            permission=CallPermission.from_dict(data["permission"]),
            actions=[
                CallPermissionAction.from_dict(action) for action in data["actions"]
            ],
        )



from __future__ import annotations

from typing import cast

from pywa.types.sent_update import *  # noqa MUST BE IMPORTED FIRST
from pywa.types.sent_update import (
    SentMessage as _SentMessage,
    SentTemplate as _SentTemplate,
    InitiatedCall as _InitiatedCall,
)
from pywa_async.types import (
    Message,
    CallbackButton,
    CallbackSelection,
    MessageStatus,
    FlowCompletion,
)
from .calls import _CallShortcutsAsync

from .. import filters as pywa_filters

__all__ = ["SentMessage", "SentTemplate", "SentTemplateStatus", "InitiatedCall"]

from pywa_async.types.base_update import _ClientShortcutsAsync


class SentMessage(_ClientShortcutsAsync, _SentMessage):
    """
    Represents a message that was sent to WhatsApp user.

    Attributes:
        id: The ID of the message.
        to_user: The user the message was sent to.
        from_phone_id: The WhatsApp ID of the sender who sent the message.
    """

    async def wait_for_reply(
        self,
        *,
        force_quote: bool = False,
        filters: pywa_filters.Filter = None,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> Message:
        """
        Wait for a message reply to the sent message.

        - Shortcut for :meth:`~pywa.client.WhatsApp.listen` with ``filters=filters.message``.

        Example:

            .. code-block:: python

                @wa.on_message(filters.command("start"))
                def start(w: WhatsApp, m: Message):
                    user_id: str = m.reply(
                        text=f"Hi {m.from_user.name}! Please enter your ID",
                        buttons=[Button(title="Cancel", callback_data="cancel")],
                    ).wait_for_reply(
                        filters=filters.text & filters.new(lambda _, m: m.text.isdigit()),
                        cancelers=filters.callback_button & filters.matches("cancel"),
                    ).text
                    ...
        Args:
            force_quote: Whether to force the reply to quote the sent message.
            filters: The filters to apply to the reply.
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for a reply.

        Returns:
            The reply message.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """
        if force_quote:
            reply_filter = pywa_filters.replays_to(self.id)
            filters = (reply_filter & filters) if filters else reply_filter
        filters = (pywa_filters.message & filters) if filters else pywa_filters.message
        return cast(
            Message,
            await self._client.listen(
                to=self.listener_identifier,
                filters=filters,
                cancelers=cancelers,
                timeout=timeout,
            ),
        )

    async def wait_until_read(
        self,
        *,
        cancel_on_new_update: bool = False,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> MessageStatus:
        """
        Wait for the message to be read by the recipient.

        - Shortcut for :meth:`~pywa.client.WhatsApp.listen` with ``filters=filters.message_status & filters.read``.

        **Note:** This method will not work if the recipient has disabled read receipts.
        make sure to use ``cancel_on_new_update=True`` to cancel the listening if the message is probably read, or use a timeout / cancelers.

        Example:

            .. code-block:: python

                @wa.on_message(filters.command("start"))
                def start(w: WhatsApp, m: Message):
                    r = m.reply("This message waits for you to read it")
                    try:
                        r.wait_until_read(cancel_on_new_update=True)
                    except ListenerCanceled as e:
                        print(e.update) # The update that canceled the listener
                        r.reply("You turned off read receipts")
                    r.reply("You read this message", quote=True)

        Args:
            cancel_on_new_update: Whether to cancel when another message/button click arrives (which may indicate the previous message was read).
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for the message to be read.

        Returns:
            The message status.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """
        if cancel_on_new_update:
            new_update_canceler = ~pywa_filters.update_id(self.id) & (
                pywa_filters.message
                | pywa_filters.callback_button
                | pywa_filters.callback_selection
            )
            cancelers = (
                (new_update_canceler & cancelers) if cancelers else new_update_canceler
            )
        return cast(
            MessageStatus,
            await self._client.listen(
                to=self.listener_identifier,
                filters=pywa_filters.message_status & pywa_filters.read,
                cancelers=cancelers,
                timeout=timeout,
            ),
        )

    async def wait_until_delivered(
        self,
        *,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> MessageStatus:
        """
        Wait for the message to be delivered to the recipient.

        Example:

            .. code-block:: python

                @wa.on_message(filters.command("start"))
                def start(w: WhatsApp, m: Message):
                    r = m.reply("This message waits for you to receive it")
                    r.wait_until_delivered()
                    r.reply("You received the message", quote=True)

        Args:
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for the message to be delivered.

        Returns:
            The message status.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """

        return cast(
            MessageStatus,
            await self._client.listen(
                to=self.listener_identifier,
                filters=pywa_filters.message_status & pywa_filters.delivered,
                cancelers=cancelers,
                timeout=timeout,
            ),
        )

    async def wait_for_click(
        self,
        *,
        filters: pywa_filters.Filter = None,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> CallbackButton:
        """
        Wait for a button click.

        Example:

            .. code-block:: python

                @wa.on_message(filters.command("start"))
                def start(w: WhatsApp, m: Message):
                    r = m.reply(
                        text="Click a button",
                        buttons=[
                            Button(title="Option 1", callback_data="option1"),
                            Button(title="Option 2", callback_data="option2"),
                        ],
                    )
                    option = r.wait_for_click()
                    r.reply(f"You clicked: {option.title}", quote=True)

        Args:
            filters: The filters to apply to the button click.
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for the button click.

        Returns:
            The clicked button.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """
        return cast(
            CallbackButton,
            await self._client.listen(
                to=self.listener_identifier,
                filters=pywa_filters.callback_button
                & (pywa_filters.replays_to(self.id) & (filters or pywa_filters.true)),
                cancelers=cancelers,
                timeout=timeout,
            ),
        )

    async def wait_for_selection(
        self,
        *,
        filters: pywa_filters.Filter = None,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> CallbackSelection:
        """
        Wait for a callback selection.

        Args:
            filters: The filters to apply to the selection.
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for the selection.

        Returns:
            The callback selection.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """

        return cast(
            CallbackSelection,
            await self._client.listen(
                to=self.listener_identifier,
                filters=pywa_filters.callback_selection
                & (pywa_filters.replays_to(self.id) & (filters or pywa_filters.true)),
                cancelers=cancelers,
                timeout=timeout,
            ),
        )

    async def wait_for_completion(
        self,
        *,
        filters: pywa_filters.Filter = None,
        cancelers: pywa_filters.Filter = None,
        timeout: float | None = None,
    ) -> FlowCompletion:
        """
        Wait for a flow completion.

        Example:

            .. code-block:: python

                @wa.on_message(filters.command("start"))
                def start(w: WhatsApp, m: Message):
                    flow_completion = m.reply(
                        text="Answer the questions",
                        buttons=FlowButton(flow_token=..., ...),
                    ).wait_for_completion()

        Args:
            filters: The filters to apply to the completion.
            cancelers: The filters to cancel the listening.
            timeout: The time to wait for the completion.

        Returns:
            The flow completion.

        Raises:
            ListenerTimeout: If the listener timed out.
            ListenerCanceled: If the listener was canceled by a filter.
            ListenerStopped: If the listener was stopped manually.
        """
        return cast(
            FlowCompletion,
            await self._client.listen(
                to=self.listener_identifier,
                filters=pywa_filters.flow_completion
                & (pywa_filters.replays_to(self.id) & (filters or pywa_filters.true)),
                cancelers=cancelers,
                timeout=timeout,
            ),
        )


class SentTemplate(SentMessage, _SentTemplate):
    """
    Represents a template message that was sent to WhatsApp user.

    Attributes:
        id: The ID of the message.
        to_user: The user the message was sent to.
        from_phone_id: The WhatsApp ID of the sender who sent the message.
        status: The status of the sent template.
    """


class InitiatedCall(_CallShortcutsAsync, _ClientShortcutsAsync, _InitiatedCall):
    """
    Represents an outgoing call initiated by the business.

    Attributes:
        id: The call ID.
        to_user: The user to whom the call was made.
        from_phone_id: The WhatsApp ID of the business phone number that initiated the call.
        success: Whether the call was successfully initiated.
    """



import pytest
from unittest.mock import Mock

from pywa import utils as pywa_utils
from pywa.server import Server, StopHandling, ContinueHandling


def test_webhook_updates_validator():
    assert pywa_utils.webhook_updates_validator(
        app_secret="1222e786b144d0e85b9f365372d93676",  # This secret key is already reset
        request_body=b'{"object":"whatsapp_business_account","entry":[{"id":"264937493375603","changes":[{"value":{"messaging_product":"whatsapp","metadata":{"display_phone_number":"15550953877","phone_number_id":"277321005464405"},"contacts":[{"profile":{"name":"PyWa Tests"},"wa_id":"972544401243"}],"messages":[{"from":"972544401243","id":"wamid.HBgMOTcyNTQ0NDAxMjQzFQIAEhggM0RFQTNCMEEwRTY3QzUwODYzMDc4NjQ4QzM4ODAxM0YA","timestamp":"1730231903","text":{"body":"Hey there! I am using PyWa."},"type":"text"}]},"field":"messages"}]}]}',
        x_hub_signature="sha256=54edfa1d7259e0eb13c677cc7d73d1b5c86cfa12433d19156e058ab9251bc441",
    )


# This private key is already reset
private_key = """
-----BEGIN ENCRYPTED PRIVATE KEY-----
MIIFJDBWBgkqhkiG9w0BBQ0wSTAxBgkqhkiG9w0BBQwwJAQQkgClH0obPRRd3cmr
s0LvRAICCAAwDAYIKoZIhvcNAgkFADAUBggqhkiG9w0DBwQIWxsUnYjosHEEggTI
tIwws0TcJNJ+IoMfNfalxb8FpV8JzIXNPKoS2vBNhni3F6tk6tYSsBW7p5Wl/1jv
0fmf6iZhGgidFT5KvxRI7Z9FyD4FAKKxMsIsrxdqblupaB/2L1dHM2foBmKZTKVP
pMkiGYCRM+uPQWao/etBvTf1IKQw5phFJ6J/NVhlU6hSh3SjMp9CN4Xy6rA7wcfu
yYfrSUBHoCZZMvdWkCxo9sQ0N/nvU1FpBO0ZZQ+WDRKKg3g8BD8jWKRVM+uzZdVx
oWLoqkIWlNuCbObYsWvK07q9qryG0WeeZtPRti6Tp/TW2NMSqIN5Zx1jHLAYPU0V
3fgt3TmHPphy8etyXEg7PBJToCgTw6snnDqQDJm4nfJPg9UxxUvkkN8Tk+5P3o/0
E2Hf3RhYIqraKhuZAToA5Isniez/L/O5Fyjye0ShOFF9auhsmLDKbdG6wL+2CX1+
cQAash4HE2Xlj9v2iVC5mlnU9e7F1EyhJNAkmbr+i2z+93Wp/XytAx9xGV6lt7J8
gv454dCSQ5brByM/D12yyB5mcKX73MkaECa2iu/qH30H10S91IgiLUOQsS9A8ABi
S32U9SlYGfwlyV7izOXiS5vnk2XiihYFKPIdoNgcq3FPDXZoJr9u/ipZl/viTzWP
2V9DTODpTWuERIUCi8Lrg82HA62yNLO6Z5ib7+oHgpsTrhxWbqvxa8T4EL7V2Hx+
AqD2/M9easd2Atl6mI8uekjndL8sKNJ2kNlG54PMddgZzoIMYPnYp0ZPCtds+KIj
pwHK0nGL+kih0R9FLHLm0YKIYgD/i13N0vD+qSd0/PC66Bcnm5CqDlOYMyweBzvv
TrnUgTCUyMVGbKJ4mpBMHQOSoVfGuE18zkfvoUmqXNEUiOj3v61ODs50zrz9GzeL
r8mehVuSY82CDqPM7vbJS+UIgrPi9cqhxv1APcV+FW+nDux3hyvyG6mHxSOoaTz/
tmAXqxZ7IJKRJW+p2Qw9FjSrQOjjfCDWRKNd3m7WhHlceKWuh6fargYbs3A7jIf7
+BaGTOfx/EcbgWrrvQ6FgdNqiijUfxmk0nTtFAaAACs/V9wOiqZLCBP+Uqv5GL7T
X2+yrJtpe2A/Jc/0VC7utOK7HPHO9y+mqkFm2SzCUnN4PDCIm6L4yDdzdjsAxj6n
YcG2AYnZhPGstAXYeQThbbgnOa76hpr4FDFb7y5OMQvbb7uDAs6goHd8yrlGdMub
KqalC3IWh9RbnDbXQLFkTG9ijaswbKu0Q4DRIjVtN/RRlYrLqKE/P/uy99GQ+zxC
vUSUci1YVbkGaTC7OMeLWf2gPptcEOXRqZhluEFIxIeMD2ykv50nfZWl6wBQRyas
+MsEoHCZYtb8ZgA8adAh0bFo/zsh9nu/HeXN6Rk0DIzBv31t6bIubsZ8VAybVetG
vkivZ9yrc5R/lNYCmIIk0YqYr4IU8GMru7l11Ojui08RTESwZ5KcsH1s9CFWuUG4
D8SsbEVopM4IbnZi0X5WWyHreEjfBrcP6+/o+3vzi+sq76v17PXlalypi4kuUjAD
78w/o5vWmS+DWkrr0DqmQZ3nFX3fcrBbP1blr1Nlb5iv3Rwy7tVMF87tpgrnk/4Y
337xpCvNLXW0EwL6KdWqJ7Y6KLjLDmeT
-----END ENCRYPTED PRIVATE KEY-----
"""


def test_default_flow_request_decryptor_encryptor():
    payload = {
        "encrypted_flow_data": "sCTmBCqjs0GkkX6n/nyZDuyjpaijuelY3I/8rlr1ZIEymEzCMnDGQdxQ9OGaKw0CEaWSgc/GLhuixa8NTQNYXAyVfTaU9H2FWEabWUb8nbZYRdYy81XHUkDCodl4SvBhhufEag==",
        "encrypted_aes_key": "gSTeWDqfKqo1eL73VstmrMm5k5lymwUwXCfuxauPFPoW7Ji9dgcG74Y6YRtoYOAch6Z/AgrR7EAlsRi/s8xT/Gx2WWz6zfcXPUQVpoIlp7EgC+HmmA2ZK64g/107yL+vKoUdL0mWJHQf1ml12HszBxOtNlW+7GAMPESNDqGpgy1R3Zgz/luStp2INtigps9w2j9+Ktp0smqxHqpUkBWp8xxoWVvzPK4H0jcFm7sjFMpiJ1e1EjApo7iDqldys0tMRC+KoOjJVD6aq1gY5s2yYL7iCXXgEAKJItTk/4/mbWWNkRtd9NoEGnMHilcjYOzlUCHehAO9fos+WCLE87JAXw==",
        "initial_vector": "5eCmDjs+VAJwdo5caZtgbw==",
    }
    decrypted_data, aes_key, iv = pywa_utils.default_flow_request_decryptor(
        encrypted_flow_data_b64=payload["encrypted_flow_data"],
        encrypted_aes_key_b64=payload["encrypted_aes_key"],
        initial_vector_b64=payload["initial_vector"],
        private_key=private_key,
        password="pywa",
    )
    assert decrypted_data == {
        "data": {},
        "flow_token": "my_flow_token",
        "screen": "",
        "action": "INIT",
        "version": "3.0",
    }

    assert (
        pywa_utils.default_flow_response_encryptor(
            response={"version": "3.0", "screen": "SUCCESS", "data": {"key": "value"}},
            aes_key=aes_key,
            iv=iv,
        )
        == "FBEoV73B8mnSt+nzfurVK704zkwHsr1uu/m953h5vNdri5G4Pe/BoDTh6SgzgjrrZ4iP12GO3kti8YW7Tn1KibKaRf8LE/gps2ATJq3nWSCI"
    )


def test_flow_request_media_decryptor():
    assert (
        pywa_utils._flow_request_media_decryptor(
            cdn_file=b"\x84a\xd2~\xd7\xe5f\xf0T$Yv\xa3\xa3\xb0\xc3\xa9>+\xca7\xd5\x81\x0cw\xe7/a\x92Q\xce\xfe\xa8>\xeaP\x915\x0b[/\x83\xd23gB\xd4\xf0\xd4c\x95w\xc6\xa2>\xb38\x7f[\xef\x87\x8d,Y\xce\xe9\xe3\xaf\x88\x04\x91e\xfc\xf9dK\\\xd0\x8e^\xa4\x88\xd7\x101JI2T\x13\xe1\xf5BR\r\x1et\xdeO@K\xbd\x95DQ\xb1(\x00\x84\xe5\xed\xe90\xf0\xd3k\xb0\xf2\x01\x88(7%\xd0\xc63<\x91\x96\x9c\xd4*\xbb.\xed\xa1D\x01D\xdd\xfa\x96eK\xd9\xe44H\x16\x0e\xb8eC\xfe\x1a\xfb\x14/\x1b\x81\x93\xef\xc5\xc2}\xe1\xd2\x9d\xca\xdf\x96\xb0V\xe96\xfcL9\x82KE\xfe\x11\xea\x8a\x9e\x13{\x19\xaba\x05\xcaW\x11\xa3#B\xa2\x93\x95&\x925~\xec\xc5Qt\x16\xed\x88\r\x16\xeb\x06$\xd8\x1c\x9d\xcc\xbec\x9f\x9e\x0bkz/\xffK\x18\x8ag(wO9t\xd7\xcc\xd7za\x94\xe3\x85a<\x8f<\xb6\x19]\xfev\x92\xa7T\xb6\x0cE\x86\x04\xfb\xdb\xb8\xb2\x93\x1d\x12\xa3>\xb1\x01vR\xde\xcb\xdf\xb2\xb3E\xdaHv\xcbg\x1bZ\xd0\x82)\xda\x14\xadL\x9b\x99\x03\n\xa9p\xda8\x11`\xc8\xfe\xeaAe\x85\xd7\x9c\r\x7f\x88\xd9 \xaf\xc8\x93\x98\xe7\xe1 Y\xc2\xc6\xc8\x9b9\x9bf\xa2\x84DL\xb7\xfc\xab\x1c\xd3yH\xb5\xedE\xc9\xddi\xd1f\xff\x18\x84\xcd\xbc\x80\x91\xa4\x04gZ\x85\x9ae\xd8\xc1Pf+\xaa=\x08\x9b~0\x16\x96\x9b\xd3\x13]\xb1Z\xca\xd4\xf8(\xf4\x14\x02\xb5\x13R\x06wY\xc4\xf3\xe6;\x10T\xca\t\xf5C\xe6)5\xf7\xef\xd1DW+\xf5\x94\xc0y\x9d\xdd\x90\x81\x1c)r}NE\x981m@\xed\x9fq\xfb^\xe0\xfaR;\xdf\x16\xce\x081\x95i={\xfeK]",
            encryption_metadata={
                "encryption_key": "202pQMDtZoAMwJwZJFVPqQOgdJRBahBmGywwSXz5tAY=",
                "hmac_key": "A/72TYylRAHTg/CdXpBtC6T6qcJ2C7Cf2qzZ/hqVASM=",
                "iv": "t1MOy02KXLbsH+NYkqkRXQ==",
                "plaintext_hash": "ZvSgxwXg5fWL7v7ggGHXtMCZYTf/nVFasOdX0p6kiP4=",
                "encrypted_hash": "pDhRHkyevzgkdg5ObY+MfzW5J6/ObZj/OrmAvyUeYA8=",
            },
        )
        == b'\xff\xd8\xff\xe0\x00\x10JFIF\x00\x01\x01\x00\x00\x01\x00\x01\x00\x00\xff\xdb\x00\x84\x00\x06\x06\x06\x06\x07\x06\x07\x08\x08\x07\n\x0b\n\x0b\n\x0f\x0e\x0c\x0c\x0e\x0f\x16\x10\x11\x10\x11\x10\x16"\x15\x19\x15\x15\x19\x15"\x1e$\x1e\x1c\x1e$\x1e6*&&*6>424>LDDL_Z_||\xa7\x01\x06\x06\x06\x06\x07\x06\x07\x08\x08\x07\n\x0b\n\x0b\n\x0f\x0e\x0c\x0c\x0e\x0f\x16\x10\x11\x10\x11\x10\x16"\x15\x19\x15\x15\x19\x15"\x1e$\x1e\x1c\x1e$\x1e6*&&*6>424>LDDL_Z_||\xa7\xff\xc2\x00\x11\x08\x00\x01\x00\x01\x03\x01"\x00\x02\x11\x01\x03\x11\x01\xff\xc4\x00\'\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x0c\x03\x01\x00\x02\x10\x03\x10\x00\x00\x02\xaa\x07\xff\xc4\x00\x02\xff\xda\x00\x0c\x03\x01\x00\x02\x00\x03\x00\x00\x00!\x03\xff\xc4\x00\x02\xff\xda\x00\x0c\x03\x01\x00\x02\x00\x03\x00\x00\x00\x10\xf3\xff\xc4\x00\x14\x11\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x02\x01\x01?\x00\x7f\xff\xc4\x00\x14\x11\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x03\x01\x01?\x00\x7f\xff\xc4\x00\x14\x10\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x01\x00\x01?\x02\x7f\xff\xc4\x00\x14\x10\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x01\x00\x01?!\x7f\xff\xc4\x00\x14\x10\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xda\x00\x08\x01\x01\x00\x01?\x10\x7f\xff\xd9'
    )


@pytest.fixture
def mock_handler(mocker):
    handler = mocker.Mock()
    handler.handle = mocker.Mock()
    return handler


@pytest.fixture
def mock_update(mocker):
    return mocker.Mock()


@pytest.fixture
def server_with_handlers(mock_handler):
    server = Server(
        None,
        "/webhook",
        None,
        None,
        None,
        None,
        None,
        "valid_token",
        None,
        None,
        None,
        None,
        None,
        True,
        False,
        False,
    )
    server._handlers = {Mock: [mock_handler]}
    return server


def test_invoke_callbacks_calls_all_handlers(
    server_with_handlers, mock_update, mock_handler
):
    server_with_handlers._invoke_callbacks(Mock, mock_update)
    mock_handler.handle.assert_called_once_with(server_with_handlers, mock_update)


def test_invoke_callbacks_stops_on_stop_handling(
    server_with_handlers, mock_update, mock_handler
):
    mock_handler.handle.side_effect = StopHandling
    server_with_handlers._invoke_callbacks(Mock, mock_update)
    mock_handler.handle.assert_called_once_with(server_with_handlers, mock_update)


def test_invoke_callbacks_continues_on_continue_handling(
    server_with_handlers, mock_update, mock_handler
):
    mock_handler.handle.side_effect = ContinueHandling
    server_with_handlers._invoke_callbacks(Mock, mock_update)
    mock_handler.handle.assert_called_once_with(server_with_handlers, mock_update)


def test_invoke_callbacks_stops_on_handled(
    server_with_handlers, mock_update, mock_handler
):
    mock_handler.handle.return_value = True
    server_with_handlers._invoke_callbacks(Mock, mock_update)
    mock_handler.handle.assert_called_once_with(server_with_handlers, mock_update)



import dataclasses
import random
from typing import Callable, TypeVar, cast

import pytest

from pywa import filters as fil
from pywa.errors import WhatsAppError, MediaUploadError
from pywa.filters import Filter
from pywa.types import (
    Message,
    CallbackSelection,
    CallbackButton,
    MessageStatus,
    ReplyToMessage,
    ReferredProduct,
)
from pywa.types.base_update import BaseUpdate
from tests.common import CLIENTS

_T = TypeVar("_T", bound=BaseUpdate)
WHATSAPP_N = "1234567890"


def same(x: _T) -> _T:
    return x


# {filename: {test_name: [(update_modifier, filter_func)]}}

FILTERS: dict[str, dict[str, list[tuple[Callable[[_T], _T], Filter]]]] = {
    "message": {
        "text": [
            (same, fil.text),
            (lambda m: modify_text(m, "hello"), fil.matches("hello")),
            (
                lambda m: modify_text(m, "hello"),
                fil.matches("hello", ignore_case=True),
            ),
            (lambda m: modify_text(m, "hi hello"), fil.contains("hello")),
            (
                lambda m: modify_text(m, "hi Hello"),
                fil.contains("hello", "Hi", ignore_case=True),
            ),
            (
                lambda m: modify_text(m, "hi bye"),
                fil.startswith("hi"),
            ),
            (
                lambda m: modify_text(m, "hi bye"),
                fil.startswith("Hi", ignore_case=True),
            ),
            (
                lambda m: modify_text(m, "hi bye"),
                fil.endswith("bye"),
            ),
            (
                lambda m: modify_text(m, "hi bye"),
                fil.endswith("Bye", ignore_case=True),
            ),
            (
                lambda m: modify_text(m, "hi bye"),
                fil.regex(r"^hi", r"bye$"),
            ),
            (
                lambda m: modify_text(m, "!start"),
                fil.command("start"),
            ),
            (
                lambda m: modify_text(m, "/start"),
                fil.command("start", prefixes="!/"),
            ),
            (
                lambda m: modify_text(m, "!Start"),
                fil.command("staRt", ignore_case=True),
            ),
            (
                lambda m: modify_text(m, "!start"),
                fil.is_command,
            ),
        ],
        "image": [
            (same, fil.image),
            (lambda m: add_caption(m, "caption"), fil.has_caption),
            (
                lambda m: modify_img_mime_type(m, "image/jpeg"),
                fil.mimetypes("image/jpeg"),
            ),
            (
                lambda m: modify_img_mime_type(m, "application/pdf"),
                fil.mimetypes("application/pdf"),
            ),
        ],
        "video": [
            (same, fil.video),
        ],
        "document": [
            (same, fil.document),
        ],
        "audio": [
            (same, fil.audio),
            (same, fil.audio_only),
        ],
        "voice": [
            (same, fil.voice),
        ],
        "sticker": [
            (same, fil.sticker),
        ],
        "static_sticker": [
            (same, fil.static_sticker),
        ],
        "animated_sticker": [
            (same, fil.animated_sticker),
        ],
        "reaction": [
            (same, fil.reaction_added),
            (
                lambda m: modify_reaction(m, "😀"),
                fil.reaction_emojis("😀"),
            ),
        ],
        "unreaction_empty": [
            (same, fil.reaction_removed),
        ],
        "unreaction_no_emoji": [(same, fil.reaction_removed)],
        "current_location": [
            (same, fil.current_location),
            (
                lambda m: modify_location(m, 37.4611794, -122.2531785),
                fil.location_in_radius(37.47, -122.25, 10),
            ),
        ],
        "contacts": [
            (same, fil.contacts),
            (
                lambda m: add_wa_number_to_contact(m),
                fil.contacts_has_wa,
            ),
        ],
        "order": [
            (same, fil.order),
        ],
        "unsupported": [(same, fil.unsupported)],
        "reply": [(same, fil.reply)],
        "replays_to": [(lambda m: modify_replies_to(m, "123"), fil.replays_to("123"))],
        "sent_to": [
            (lambda m: modify_send_to(m, "123"), fil.sent_to(phone_number_id="123"))
        ],
        "sent_to_me": [
            (lambda m: modify_send_to(m, CLIENTS.keys()[0].phone_id), fil.sent_to_me)
        ],
        "from_users": [(lambda m: modify_send_from(m, "123"), fil.from_users("123"))],
        "from_countries": [
            (lambda m: modify_send_from(m, "97212345678"), fil.from_countries("972"))
        ],
        "has_referred_product": [
            (lambda m: modify_referred_product(m, "IPHONE"), fil.has_referred_product)
        ],
        "forwarded": [(same, fil.forwarded)],
        "forwarded_many_times": [(same, fil.forwarded)],
        "interactive_message_with_err": [],
    },
    "callback_button": {
        "button": [
            (
                lambda m: modify_callback_data(m, "hi"),
                fil.matches("hi"),
            ),
            (
                lambda m: modify_callback_data(m, "Hi"),
                fil.matches("hi", ignore_case=True),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.contains("hi"),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.contains("Hi", ignore_case=True),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.startswith("hi"),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.startswith("Hi", ignore_case=True),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.endswith("bye"),
            ),
            (
                lambda m: modify_callback_data(m, "hi bye"),
                fil.endswith("Bye", ignore_case=True),
            ),
            (
                lambda m: modify_callback_data(m, "data:123"),
                fil.regex("^data:", r"\d{3}$"),
            ),
        ],
        "quick_reply": [],
    },
    "callback_selection": {
        "callback": [],
        "description": [],
    },
    "message_status": {
        "sent": [(same, fil.sent)],
        "delivered": [(same, fil.delivered)],
        "read": [(same, fil.read)],
        "failed": [
            (
                lambda m: modify_status_err(
                    m, WhatsAppError.from_dict({"code": 131053, "message": "error"})
                ),
                fil.failed_with(MediaUploadError),
            ),
            (
                lambda m: modify_status_err(
                    m, WhatsAppError.from_dict({"code": 131053, "message": "error"})
                ),
                fil.failed_with(131053),
            ),
        ],
        "with_tracker": [
            (same, fil.with_tracker),
            (lambda s: modify_status_tracker(s, "tracker"), fil.matches("tracker")),
        ],
    },
    "template_status": {
        "approved": [(same, fil.template_status)],
    },
    "flow_completion": {
        "completion": [
            (same, fil.flow_completion),
        ]
    },
    "chat_opened": {"chat_opened": [(same, fil.chat_opened)]},
}


def test_combinations():
    true, false = fil.new(lambda _, __: True), fil.new(lambda _, __: False)
    assert true.check_sync(None, None)
    assert not false.check_sync(None, None)
    assert (true & true).check_sync(None, None)
    assert not (true & false).check_sync(None, None)
    assert (true | false).check_sync(None, None)
    assert not (false | false).check_sync(None, None)


@pytest.mark.asyncio
async def test_combinations_async():
    async def _async_true(_, __):
        return True

    async def _async_false(_, __):
        return False

    async_true, async_false = fil.new(_async_true), fil.new(_async_false)
    sync_true, sync_false = fil.new(lambda _, __: True), fil.new(lambda _, __: False)

    assert await async_true.check_async(None, None)
    assert not await async_false.check_async(None, None)
    assert await (async_true & async_true).check_async(None, None)
    assert await (sync_true & async_true).check_async(None, None)
    assert await (async_true & sync_true).check_async(None, None)
    assert not await (async_true & async_false).check_async(None, None)
    assert not await (sync_true & async_false).check_async(None, None)
    assert await (async_true | async_false).check_async(None, None)
    assert await (sync_true | async_false).check_async(None, None)
    assert await (async_false | sync_true).check_async(None, None)
    assert not await (async_false | async_false).check_async(None, None)
    assert not await (sync_false | sync_false).check_async(None, None)


def test_filters():
    for client, update_files in CLIENTS.items():
        for file, updates in update_files.items():
            for update_name, update in updates.items():
                for update_modifier, filter_obj in FILTERS.get(file.stem, {}).get(
                    update_name, ()
                ):
                    modified_update = update_modifier(update)
                    try:
                        assert cast(Filter, filter_obj).check_sync(
                            client, modified_update
                        )
                    except AssertionError as e:
                        raise AssertionError(
                            f"Test {file.stem}/{update_name} failed on {modified_update}"
                        ) from e


def modify_text(msg: Message, to: str):
    return dataclasses.replace(msg, text=to)


def add_caption(msg: Message, caption: str):
    return dataclasses.replace(msg, caption=caption)


def modify_img_mime_type(msg: Message, mime_type: str):
    return dataclasses.replace(
        msg, image=dataclasses.replace(msg.image, mime_type=mime_type)
    )


def modify_reaction(msg: Message, emoji: str | None):
    return dataclasses.replace(
        msg, reaction=dataclasses.replace(msg.reaction, emoji=emoji)
    )


def modify_location(msg: Message, lat: float, lon: float):
    return dataclasses.replace(
        msg, location=dataclasses.replace(msg.location, latitude=lat, longitude=lon)
    )


def add_wa_number_to_contact(msg: Message):
    return dataclasses.replace(
        msg,
        contacts=(
            dataclasses.replace(
                msg.contacts[0],
                phones=[
                    dataclasses.replace(msg.contacts[0].phones[0], wa_id="123456789")
                ],
            ),
        ),
    )


def modify_callback_data(clb: CallbackButton | CallbackSelection, data: str):
    return dataclasses.replace(clb, data=data)


def modify_status_err(status: MessageStatus, err: WhatsAppError):
    return dataclasses.replace(status, error=err)


def modify_send_from(msg: Message, wa_id: str):
    return dataclasses.replace(
        msg, from_user=dataclasses.replace(msg.from_user, wa_id=wa_id)
    )


def modify_send_to(msg: Message, wa_id: str):
    return dataclasses.replace(
        msg, metadata=dataclasses.replace(msg.metadata, phone_number_id=wa_id)
    )


def modify_replies_to(msg: Message, message_id: str):
    return dataclasses.replace(
        msg,
        reply_to_message=dataclasses.replace(
            msg.reply_to_message, message_id=message_id
        ),
    )


def modify_referred_product(msg: Message, product: str):
    reply_to_msg = ReplyToMessage(
        message_id=msg.id,
        from_user_id=msg.from_user.wa_id,
        referred_product=ReferredProduct(catalog_id="123", sku=product),
    )
    return dataclasses.replace(msg, reply_to_message=reply_to_msg)


def modify_status_tracker(s: MessageStatus, tracker: str):
    return dataclasses.replace(s, tracker=tracker)



"""
Usefully filters to use in your handlers.

>>> from pywa import filters as fil
>>> from pywa import WhatsApp, types
>>> wa = WhatsApp(...)

>>> @wa.on_message(fil.text.command("start"))
... def on_hi_msg(_: WhatsApp, m: types.Message):
...     print("This is a welcome message!")
"""

from __future__ import annotations

__all__ = [
    "new",
    "update_id",
    "forwarded",
    "forwarded_many_times",
    "reply",
    "replays_to",
    "has_referred_product",
    "sent_to",
    "sent_to_me",
    "from_users",
    "from_countries",
    "matches",
    "contains",
    "startswith",
    "endswith",
    "regex",
    "message",
    "text",
    "is_command",
    "command",
    "media",
    "mimetypes",
    "extensions",
    "has_caption",
    "image",
    "video",
    "audio",
    "audio_only",
    "voice",
    "document",
    "sticker",
    "animated_sticker",
    "static_sticker",
    "reaction",
    "reaction_added",
    "reaction_removed",
    "reaction_emojis",
    "unsupported",
    "location",
    "current_location",
    "location_in_radius",
    "contacts",
    "contacts_has_wa",
    "order",
    "callback_button",
    "callback_selection",
    "message_status",
    "sent",
    "delivered",
    "read",
    "failed",
    "failed_with",
    "with_tracker",
    "flow_completion",
    "template_status",
    "template_status_approved",
    "template_status_rejected",
    "template_quality",
    "template_category",
    "template_components",
    "chat_opened",
    "call_connect",
    "outgoing_call",
    "incoming_call",
    "call_status",
    "call_answered",
    "call_rejected",
    "call_ringing",
    "call_permission_update",
    "call_permission_accepted",
    "call_permission_rejected",
    "call_terminate",
    "phone_number_change",
    "identity_change",
]

import re
from typing import TYPE_CHECKING, Callable, Iterable, TypeVar, Awaitable, Any

from .errors import ReEngagementMessage, WhatsAppError
from . import utils
from .types import Message as _Msg
from .types import CallbackButton as _Clb
from .types import CallbackSelection as _Cls
from .types import MessageStatus as _Ms
from .types import MessageStatusType as _Mst
from .types import MessageType as _Mt
from .types import TemplateStatusUpdate as _Ts
from .types import TemplateQualityUpdate as _Tq
from .types import TemplateCategoryUpdate as _Tc
from .types import TemplateComponentsUpdate as _Tcc
from .types import FlowCompletion as _Fc
from .types import ChatOpened as _Co
from .types import CallConnect as _Cc
from .types import CallStatus as _Cst
from .types import CallPermissionUpdate as _Cpu
from .types import CallTerminate as _Ct
from .types import PhoneNumberChange as _Pnc
from .types import IdentityChange as _Ic

from .types.base_update import (
    BaseUpdate as _BaseUpdate,
)  # noqa
from .types.calls import CallDirection, CallPermissionResponse, CallStatusType
from .types.templates import TemplateStatus

if TYPE_CHECKING:
    from pywa import WhatsApp as _Wa


_T = TypeVar("_T", bound=_BaseUpdate)


class Filter:
    """Base filter class handling both sync and async."""

    def check_sync(self, wa: _Wa, update: Any) -> bool:
        raise NotImplementedError

    async def check_async(self, wa: _Wa, update: Any) -> bool:
        raise NotImplementedError

    def has_async(self) -> bool:
        raise NotImplementedError

    def __and__(self, other: Filter) -> Filter:
        return AndFilter(self, other)

    def __or__(self, other: Filter) -> Filter:
        return OrFilter(self, other)

    def __invert__(self) -> Filter:
        return NotFilter(self)


class AndFilter(Filter):
    def __init__(self, left: Filter, right: Filter):
        self.left = left
        self.right = right

    def check_sync(self, wa: _Wa, update: _T) -> bool:
        return self.left.check_sync(wa, update) and self.right.check_sync(wa, update)

    async def check_async(self, wa: _Wa, update: _T) -> bool:
        return await self.left.check_async(wa, update) and await self.right.check_async(
            wa, update
        )

    def has_async(self) -> bool:
        return self.left.has_async() or self.right.has_async()


class OrFilter(Filter):
    def __init__(self, left: Filter, right: Filter):
        self.left = left
        self.right = right

    def check_sync(self, wa: _Wa, update: _T) -> bool:
        return self.left.check_sync(wa, update) or self.right.check_sync(wa, update)

    async def check_async(self, wa: _Wa, update: _T) -> bool:
        return await self.left.check_async(wa, update) or await self.right.check_async(
            wa, update
        )

    def has_async(self) -> bool:
        return self.left.has_async() or self.right.has_async()


class NotFilter(Filter):
    def __init__(self, fil: Filter):
        self.filter = fil

    def check_sync(self, wa: _Wa, update: _T) -> bool:
        return not self.filter.check_sync(wa, update)

    async def check_async(self, wa: _Wa, update: _T) -> bool:
        return not await self.filter.check_async(wa, update)

    def has_async(self) -> bool:
        return self.filter.has_async()


def new(
    func: Callable[[_Wa, _T], bool | Awaitable[bool]], name: str | None = None
) -> Filter:
    """Factory function to create a filter from a function (sync or async)."""

    is_async = utils.is_async_callable(func)

    def check_sync(self, wa: _Wa, update: _T) -> bool:
        return func(wa, update)

    async def check_async(self, wa: _Wa, update: _T) -> bool:
        if is_async:
            return await func(wa, update)
        return func(wa, update)

    def has_async(self) -> bool:
        return is_async

    return type(
        name or func.__name__ or Filter,
        (Filter,),
        {
            "check_sync": check_sync,
            "check_async": check_async,
            "has_async": has_async,
        },
    )()


true = new(lambda _, __: True, name="true")
"""Filter that always returns True."""

false = new(lambda _, __: False, name="false")
"""Filter that always returns False."""

forwarded = new(
    lambda _, m: m.forwarded,
    name="forwarded",
)
"""
Filter for forwarded messages.

>>> filters.forwarded
"""

forwarded_many_times = new(
    lambda _, m: m.forwarded_many_times, name="forwarded_many_times"
)
"""
Filter for messages that have been forwarded many times.

>>> filters.forwarded_many_times
"""

reply = new(lambda _, m: m.reply_to_message is not None, name="reply")
"""
Filter for messages that reply to another message.

>>> filters.reply
"""


def update_id(id_: str) -> Filter:
    """
    Filter for updates that have the given id.

    >>> update_id("wamid.HBKHUIyNTM4NjAfiefhwojfMTNFQ0Q2MERGRjVDMUHUIGGA=")
    """
    return new(lambda _, u: u.id == id_, name="update_id")


def replays_to(*msg_ids: str) -> Filter:
    """
    Filter for messages that reply to any of the given message ids.

    >>> replays_to("wamid.HBKHUIyNTM4NjAfiefhwojfMTNFQ0Q2MERGRjVDMUHUIGGA=")
    """
    return new(
        lambda _, m: m.reply_to_message is not None
        and m.reply_to_message.message_id in msg_ids
    )


has_referred_product = new(
    lambda _, m: (
        m.reply_to_message is not None
        and m.reply_to_message.referred_product is not None
    ),
    name="has_referred_product",
)
"""
Filter for messages that user sends to ask about a product

>>> filters.referred_product
"""


def sent_to(*, display_phone_number: str = None, phone_number_id: str = None) -> Filter:
    """
    Filter for updates that are sent to the given phone number.

    - Use this filter when you choose not filter updates (e.g. ``WhatsApp(..., filter_updates=False)``) so you can still filter for messages that are sent to specific phone numbers.


    >>> sent_to(display_phone_number="+1 555-555-5555")
    >>> sent_to(phone_number_id="123456789")
    """
    if not (display_phone_number or phone_number_id):
        raise ValueError(
            "You must provide either display_phone_number or phone_number_id"
        )
    return new(
        lambda _, m: (
            m.metadata.display_phone_number == display_phone_number
            if display_phone_number
            else m.metadata.phone_number_id == phone_number_id
        ),
        name="sent_to",
    )


sent_to_me = new(
    lambda wa, m: sent_to(phone_number_id=wa.phone_id).check_sync(wa, m),
    name="sent_to_me",
)
"""
Filter for updates that are sent to the client phone number.

- Use this filter when you choose not filter updates (e.g. ``WhatsApp(..., filter_updates=False)``) so you can still filter for messages that are sent to the client phone number.

>>> sent_to_me
"""


def from_users(
    *numbers: str,
) -> Filter:
    """
    Filter for updates that are sent from the given numbers.

    >>> from_users("+1 555-555-5555", "972123456789")
    """
    only_nums_pattern = re.compile(r"\D")
    numbers = tuple(re.sub(only_nums_pattern, "", n) for n in numbers)
    return new(lambda _, m: m.from_user.wa_id in numbers, name="from_users")


def from_countries(
    *prefixes: str | int,
) -> Filter:
    """
    Filter for updates that are sent from the given country codes.

    - See https://countrycode.org/ for a list of country codes.

    It is always recommended to restrict the countries that can use your bot. remember that you pay for
    every conversation that you reply to.

    >>> from_countries("972", "1") # Israel and USA
    """
    codes = tuple(str(p) for p in prefixes)
    return new(lambda _, m: m.from_user.wa_id.startswith(codes), name="from_countries")


def matches(*strings: str, ignore_case: bool = False) -> Filter:
    """
    Filter for messages that are matching (``==``) any of the given strings.

    The strings will be checked against the following fields:
        - :class:`Message`: ``text``, ``caption``
        - :class:`CallbackButton`: ``data``
        - :class:`CallbackSelection`: ``data``
        - :class:`MessageStatus`: ``tracker``
        - :class:`FlowCompletion`: ``token``, ``body``

    >>> matches("Hello", "Hi")

    Args:
        *strings: The strings to match.
        ignore_case: Whether to ignore case when matching.
    """
    strings = tuple(m.lower() for m in strings) if ignore_case else strings
    return new(
        lambda _, m: any(
            (txt.lower() if ignore_case else txt) in strings
            for txt_field in m._txt_fields
            if (txt := getattr(m, txt_field)) is not None
        )
        if getattr(m, "_txt_fields", None)
        else False,
        name="matches",
    )


def startswith(*prefixes: str, ignore_case: bool = False) -> Filter:
    """
    Filter for updates that start with any of the given prefixes.

    The prefixes will be checked against the following fields:
        - :class:`Message`: ``text``, ``caption``
        - :class:`CallbackButton`: ``data``
        - :class:`CallbackSelection`: ``data``
        - :class:`MessageStatus`: ``tracker``
        - :class:`FlowCompletion`: ``token``, ``body``

    >>> startswith("Hello", "Hi", ignore_case=True)

    Args:
        *prefixes: The prefixes to match.
        ignore_case: Whether to ignore case when matching.
    """
    prefixes = tuple(m.lower() for m in prefixes) if ignore_case else prefixes
    return new(
        lambda _, u: any(
            (txt.lower() if ignore_case else txt).startswith(prefixes)
            for txt_field in u._txt_fields
            if (txt := getattr(u, txt_field)) is not None
        )
        if getattr(u, "_txt_fields", None)
        else False,
        name="startswith",
    )


def endswith(*suffixes: str, ignore_case: bool = False) -> Filter:
    """
    Filter for updates that end with any of the given suffixes.

    The suffixes will be checked against the following fields:
        - :class:`Message`: ``text``, ``caption``
        - :class:`CallbackButton`: ``data``
        - :class:`CallbackSelection`: ``data``
        - :class:`MessageStatus`: ``tracker``
        - :class:`FlowCompletion`: ``token``, ``body``

    >>> endswith("Hello", "Hi", ignore_case=True)

    Args:
        *suffixes: The suffixes to match.
        ignore_case: Whether to ignore case when matching.
    """
    suffixes = tuple(m.lower() for m in suffixes) if ignore_case else suffixes
    return new(
        lambda _, u: any(
            (txt.lower() if ignore_case else txt).endswith(suffixes)
            for txt_field in u._txt_fields
            if (txt := getattr(u, txt_field)) is not None
        )
        if getattr(u, "_txt_fields", None)
        else False,
        name="endswith",
    )


def contains(*words: str, ignore_case: bool = False) -> Filter:
    """
    Filter for updates that contain any of the given words.

    The words will be checked against the following fields:
        - :class:`Message`: ``text``, ``caption``
        - :class:`CallbackButton`: ``data``
        - :class:`CallbackSelection`: ``data``
        - :class:`MessageStatus`: ``tracker``
        - :class:`FlowCompletion`: ``token``, ``body``

    >>> contains("Hello", "Hi", ignore_case=True)

    Args:
        *words: The words to match.
        ignore_case: Whether to ignore case when matching.
    """
    words = tuple(m.lower() for m in words) if ignore_case else words
    return new(
        lambda _, u: any(
            word in (txt.lower() if ignore_case else txt)
            for word in words
            for txt_field in u._txt_fields
            if (txt := getattr(u, txt_field)) is not None
        )
        if getattr(u, "_txt_fields", None)
        else False,
        name="contains",
    )


def regex(*patterns: str | re.Pattern, flags: int = 0) -> Filter:
    """
    Filter for updates that match any of the given regex patterns.

    The patterns will be checked against the following fields:
        - :class:`Message`: ``text``, ``caption``
        - :class:`CallbackButton`: ``data``
        - :class:`CallbackSelection`: ``data``
        - :class:`MessageStatus`: ``tracker``
        - :class:`FlowCompletion`: ``token``, ``body``

    >>> regex(r"Hello|Hi")

    Args:
        *patterns: The regex patterns to match.
        flags: The regex flags to use.
    """
    patterns = tuple(
        p if isinstance(p, re.Pattern) else re.compile(p, flags) for p in patterns
    )
    return new(
        lambda _, u: any(
            re.match(p, txt)
            for p in patterns
            for txt_field in u._txt_fields
            if (txt := getattr(u, txt_field)) is not None
        )
        if getattr(u, "_txt_fields", None)
        else False,
        name="regex",
    )


message = new(lambda _, m: isinstance(m, _Msg), name="message")
"""Filter for all messages."""


def mimetypes(*mmtps: str) -> Filter:
    """
    Filter for media messages that match any of the given mime types.

    - `Supported Media Types on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/reference/media#supported-media-types>`_.

    >>> mimetypes("application/pdf", "image/png")
    """
    return new(
        lambda _, m: m.media.mime_type in mmtps,
        name="mimetypes",
    )


def extensions(*exts: str) -> Filter:
    """
    Filter for media messages that match any of the given extensions.

    >>> extensions(".pdf", ".png")
    """
    return new(
        lambda _, m: m.media.extension in exts,
        name="extensions",
    )


media = new(lambda _, m: m.has_media, name="media")
"""Filter for media messages (images, videos, documents, audio, stickers)."""

is_command = new(
    lambda _, m: m.type == _Mt.TEXT and m.text.startswith(("/", "!")),
    name="is_command",
)
"""
Filter for text messages that are commands (start with ``/`` or ``!``).

>>> filters.is_command
"""


def command(
    *cmds: str,
    prefixes: str | Iterable[str] = "/!",
    ignore_case: bool = False,
) -> Filter:
    """
    Filter for text messages that are commands.


    >>> command("start", "hello", prefixes="/", ignore_case=True)

    Args:
        *cmds: The command/s to filter for (e.g. "start", "hello").
        prefixes: The prefix/s to filter for (default: "/!", i.e. "/start").
        ignore_case: Whether to ignore case when matching (default: ``False``).
    """
    cmds = tuple(c.lower() for c in cmds) if ignore_case else cmds
    return new(
        lambda _, m: m.type == _Mt.TEXT
        and (
            m.text[0] in prefixes
            and (m.text[1:].lower() if ignore_case else m.text[1:]).startswith(cmds)
        ),
        name="command",
    )


text = new(lambda _, m: m.type == _Mt.TEXT, name="text")
"""Filter for text messages."""

has_caption = new(
    lambda _, m: m.caption is not None,
    name="media_has_caption",
)
"""Filter for media messages that have a caption."""

image = new(lambda _, m: m.type == _Mt.IMAGE, name="image")
"""Filter for image messages."""

video = new(lambda _, m: m.type == _Mt.VIDEO, name="video")
"""Filter for video messages."""

document = new(lambda _, m: m.type == _Mt.DOCUMENT, name="document")
"""Filter for document messages."""

audio = new(lambda _, m: m.type == _Mt.AUDIO, name="audio")
"""Filter for audio messages (both voice notes and audio files)."""

audio_only = new(
    lambda _, m: m.type == _Mt.AUDIO and not m.audio.voice, name="audio_only"
)
"""Filter for audio messages that are not voice notes."""

voice = new(lambda _, m: m.type == _Mt.AUDIO and m.audio.voice, name="voice")
"""Filter for audio messages that are voice notes."""

sticker = new(lambda _, m: m.type == _Mt.STICKER, name="sticker")
"""Filter for sticker messages (both static and animated)."""

animated_sticker = new(
    lambda _, m: m.type == _Mt.STICKER and m.sticker.animated, name="animated_sticker"
)
"""Filter for animated sticker messages."""

static_sticker = new(
    lambda _, m: m.type == _Mt.STICKER and not m.sticker.animated, name="static_sticker"
)
"""Filter for static sticker messages."""

location = new(lambda _, m: m.type == _Mt.LOCATION, name="location")
"""Filter for location messages."""

current_location = new(
    lambda _, m: m.type == _Mt.LOCATION and m.location.current_location,
    name="current_location",
)
"""Filter for location messages that are current locations."""


def location_in_radius(lat: float, lon: float, radius: float | int) -> Filter:
    """
    Filter for location messages that are in a given radius.

    >>> location_in_radius(lat=37.48508108998884, lon=-122.14744733542707, radius=1)

    Args:
        lat: Latitude of the center of the radius.
        lon: Longitude of the center of the radius.
        radius: Radius in kilometers.
    """

    return new(
        lambda _, m: m.type == _Mt.LOCATION
        and m.location.in_radius(lat=lat, lon=lon, radius=radius),
        name="location_in_radius",
    )


reaction = new(lambda _, m: m.type == _Mt.REACTION, name="reaction")
"""Filter for reaction messages (both added and removed)."""


reaction_added = new(
    lambda _, m: m.type == _Mt.REACTION and m.reaction.emoji is not None,
    name="reaction_added",
)
"""Filter for reaction messages that were added to a message."""


reaction_removed = new(
    lambda _, m: m.type == _Mt.REACTION and m.reaction.emoji is None,
    name="reaction_removed",
)
"""Filter for reaction messages that were removed from a message."""


def reaction_emojis(*emojis: str) -> Filter:
    """
    Filter for custom reaction messages. pass emojis as strings.

    >>> reaction_emojis("👍","👎")
    """
    return new(
        lambda _, m: m.type == _Mt.REACTION and m.reaction.emoji in emojis,
        name="reaction_emojis",
    )


contacts = new(lambda _, m: m.type == _Mt.CONTACTS, name="contacts")
"""Filter for contacts messages."""


contacts_has_wa = new(
    lambda _, m: m.type == _Mt.CONTACTS
    and (
        any(
            (
                p.wa_id
                for p in (phone for contact in m.contacts for phone in contact.phones)
            )
        )
    ),
    name="contacts_any_has_wa",
)
"""Filter for contacts messages that have a WhatsApp account."""


order = new(lambda _, m: m.type == _Mt.ORDER, name="order")
"""Filter for order messages."""


unsupported = new(lambda _, m: m.type == _Mt.UNSUPPORTED, name="unsupported")
"""Filter for all unsupported messages."""


callback_button = new(lambda _, c: isinstance(c, _Clb), name="callback_button")
"""Filter for callback buttons."""

callback_selection = new(lambda _, c: isinstance(c, _Cls), name="callback_selection")
"""Filter for callback selections."""

message_status = new(lambda _, s: isinstance(s, _Ms), name="message_status")

sent = new(lambda _, s: s.status == _Mst.SENT, name="status_sent")
"""Filter for messages that have been sent."""

delivered = new(lambda _, s: s.status == _Mst.DELIVERED, name="status_delivered")
"""Filter for messages that have been delivered."""

read = new(lambda _, s: s.status == _Mst.READ, name="status_read")
"""Filter for messages that have been read."""

failed = new(lambda _, s: s.status == _Mst.FAILED, name="status_failed")
"""Filter for status updates of messages that have failed to send."""


def failed_with(
    *errors: type[WhatsAppError] | int,
) -> Filter:
    """
    Filter for status updates of messages that have failed to send with the given error/s.

    Args:
        *errors: The exceptions from :mod:`pywa.errors` or error codes to match.

    >>> failed_with(ReEngagementMessage)
    >>> failed_with(131051)
    """
    error_codes = tuple(c for c in errors if isinstance(c, int))
    exceptions = tuple(
        e for e in errors if e not in error_codes and issubclass(e, WhatsAppError)
    )
    return new(
        lambda _, s: s.status == _Mst.FAILED
        and (
            any((isinstance(s.error, e) for e in exceptions))
            or s.error.code in error_codes
        ),
        name="status_failed_with",
    )


with_tracker = new(lambda _, s: s.tracker is not None, name="with_tracker")
"""Filter for status updates that have a tracker."""

template_status = new(lambda _, s: isinstance(s, _Ts), name="template_status")
"""Filters for template status updates."""

template_status_approved = new(
    lambda _, s: s.status == TemplateStatus.APPROVED, name="template_status_approved"
)

template_status_rejected = new(
    lambda _, s: s.status == TemplateStatus.REJECTED, name="template_status_rejected"
)

template_quality = new(lambda _, s: isinstance(s, _Tq), name="template_quality")
"""Filters for template quality updates."""

template_category = new(lambda _, s: isinstance(s, _Tc), name="template_category")
"""Filters for template category updates."""

template_components = new(lambda _, s: isinstance(s, _Tcc), name="template_components")
"""Filters for template components updates."""

flow_completion = new(lambda _, f: isinstance(f, _Fc), name="flow_completion")
"""Filter for flow completion updates."""

chat_opened = new(lambda _, c: isinstance(c, _Co), name="chat_opened")
"""Filter for chat opened updates."""

call_connect = new(lambda _, c: isinstance(c, _Cc), name="call_connect")
"""Filter for call connect updates."""

outgoing_call = new(
    lambda _, c: c.direction == CallDirection.BUSINESS_INITIATED,
    name="outgoing_call",
)
incoming_call = new(
    lambda _, c: c.direction == CallDirection.USER_INITIATED,
    name="incoming_call",
)
"""Filter for incoming call updates."""

call_status = new(lambda _, c: isinstance(c, _Cst), name="call_status")
"""Filter for call status updates."""

call_answered = new(
    lambda _, c: c.status == CallStatusType.ACCEPTED,
    name="call_answered",
)
call_rejected = new(
    lambda _, c: c.status == CallStatusType.REJECTED, name="call_rejected"
)
call_ringing = new(lambda _, c: c.status == CallStatusType.RINGING, name="call_ringing")

call_permission_update = new(
    lambda _, c: isinstance(c, _Cpu), name="call_permission_update"
)
"""Filter for call permission updates."""

call_permission_accepted = new(
    lambda _, c: c.response == CallPermissionResponse.ACCEPT,
    name="call_permission_accepted",
)
call_permission_rejected = new(
    lambda _, c: c.response == CallPermissionResponse.REJECT,
    name="call_permission_rejected",
)

call_terminate = new(lambda _, c: isinstance(c, _Ct), name="call_terminate")
"""Filter for call terminate updates."""

phone_number_change = new(lambda _, c: isinstance(c, _Pnc), name="phone_number_change")
"""Filter for phone number change updates."""

identity_change = new(lambda _, c: isinstance(c, _Ic), name="identity_change")
"""Filter for identity change updates."""



import asyncio
import copy
import logging
from typing import Callable, cast, TYPE_CHECKING

from pywa.types.base_update import BaseUpdate, BaseUserUpdate
from . import errors, handlers
from . import utils
from .handlers import (
    MessageHandler,
    MessageStatusHandler,
    CallbackButtonHandler,
    CallbackSelectionHandler,
    FlowCompletionHandler,
    RawUpdateHandler,
    Handler,
    TemplateStatusUpdateHandler,
    ChatOpenedHandler,
    EncryptedFlowRequestType,
)
from .types import (
    Message,
)
from .types import (
    TemplateStatusUpdate,
    MessageStatus,
    CallbackButton,
    CallbackSelection,
    ChatOpened,
    FlowCompletion,
    ContinueHandling,
    StopHandling,
)

_logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from pywa_async import WhatsApp


class Server:
    async def webhook_challenge_handler(
        self: "WhatsApp", vt: str, ch: str
    ) -> tuple[str, int]:
        """
        Handle the verification challenge from the webhook manually.

        - Use this function only if you are using a custom server (e.g. Django etc.).

        Args:
            vt: The verify token param (utils.HUB_VT).
            ch: The challenge param (utils.HUB_CH).

        Returns:
            A tuple containing the challenge and the status code.
        """
        return super().webhook_challenge_handler(vt=vt, ch=ch)

    async def webhook_update_handler(
        self: "WhatsApp", update: bytes, hmac_header: str = None
    ) -> tuple[str, int]:
        """
        Handle the incoming update from the webhook manually.

        - Use this function only if you are using a custom server (e.g. Django etc.).

        Args:
            update: The incoming raw update from the webhook (bytes)
            hmac_header: The ``X-Hub-Signature-256`` header (to validate the signature, use ``utils.HUB_SIG`` for the key).

        Returns:
            A tuple containing the response and the status code.
        """
        res, status, update_dict, update_hash = self._check_and_prepare_update(
            update=update, hmac_header=hmac_header
        )
        if res:
            return res, status
        await self._call_handlers(update_dict)
        return self._after_handling_update(update_hash)

    def _register_routes(self: "WhatsApp") -> None:
        match self._server_type:
            case utils.ServerType.FLASK:
                import flask

                if not utils.is_installed("asgiref"):  # flask[async]
                    raise ValueError(
                        "Flask with ASGI is required to handle incoming updates asynchronously. Please install "
                        """the `asgiref` package (`pip install "flask[async]"` / `pip install "asgiref"`)"""
                    )

                _logger.info("Using Flask with ASGI")

                @self._server.route(self._webhook_endpoint, methods=["GET"])
                @utils.rename_func(f"('{self._webhook_endpoint}')")
                async def pywa_challenge() -> flask.Response:
                    """Automatically generated by pywa to handle the verification challenge."""
                    ch, code = await self.webhook_challenge_handler(
                        vt=flask.request.args.get(utils.HUB_VT),
                        ch=flask.request.args.get(utils.HUB_CH),
                    )
                    return flask.Response(
                        response=ch,
                        status=code,
                        content_type="text/plain",
                        headers={
                            "X-Content-Type-Options": "nosniff",
                        },
                    )

                @self._server.route(self._webhook_endpoint, methods=["POST"])
                @utils.rename_func(f"('{self._webhook_endpoint}')")
                async def pywa_webhook() -> flask.Response:
                    """Automatically generated by pywa to handle incoming updates."""
                    res, status = await self.webhook_update_handler(
                        update=flask.request.data,
                        hmac_header=flask.request.headers.get(utils.HUB_SIG),
                    )
                    return flask.Response(
                        response=res,
                        status=status,
                        content_type="text/plain",
                        headers={
                            "X-Content-Type-Options": "nosniff",
                        },
                    )

            case utils.ServerType.FASTAPI:
                _logger.info("Using FastAPI")
                import fastapi

                @self._server.get(self._webhook_endpoint)
                @utils.rename_func(f"('{self._webhook_endpoint}')")
                async def pywa_challenge(
                    vt: str = fastapi.Query(alias=utils.HUB_VT, example="xyzxyz"),
                    ch: str = fastapi.Query(alias=utils.HUB_CH, example="1858252904"),
                ) -> fastapi.Response:
                    """Automatically generated by pywa to handle the verification challenge."""
                    content, status_code = await self.webhook_challenge_handler(
                        vt=vt,
                        ch=ch,
                    )
                    return fastapi.Response(
                        content=content,
                        status_code=status_code,
                        media_type="text/plain",
                        headers={
                            "X-Content-Type-Options": "nosniff",
                        },
                    )

                @self._server.post(self._webhook_endpoint)
                @utils.rename_func(f"('{self._webhook_endpoint}')")
                async def pywa_webhook(
                    req: fastapi.Request,
                    hmac_header: str = fastapi.Header(
                        alias=utils.HUB_SIG, example="sha256=..."
                    ),
                ) -> fastapi.Response:
                    """Automatically generated by pywa to handle incoming updates."""
                    content, status_code = await self.webhook_update_handler(
                        update=await req.body(),
                        hmac_header=hmac_header,
                    )
                    return fastapi.Response(
                        content=content,
                        status_code=status_code,
                        media_type="text/plain",
                        headers={
                            "X-Content-Type-Options": "nosniff",
                        },
                    )
            case None:
                _logger.info("Using a custom server")

            case _:
                raise ValueError(
                    f"The `server` must be one of {utils.ServerType.protocols_names()} or None for a custom server"
                )

    def _register_flow_callback_wrapper(
        self: "WhatsApp",
        callback_wrapper: handlers.FlowRequestCallbackWrapper,
    ) -> handlers.FlowRequestCallbackWrapper:
        """Register the flow callback wrapper to the server."""
        match self._server_type:
            case utils.ServerType.FLASK:
                if not utils.is_installed("asgiref"):  # flask[async]
                    raise ValueError(
                        "Flask with ASGI is required to handle incoming updates asynchronously. Please install "
                        """the `asgiref` package (`pip install "flask[async]"` / `pip install "asgiref"`)"""
                    )

                import flask

                @self._server.route(callback_wrapper._endpoint, methods=["POST"])
                @utils.rename_func(f"('{callback_wrapper._endpoint}')")
                async def pywa_flow() -> tuple[str, int]:
                    """Automatically generated by pywa to handle incoming flow requests."""
                    return await callback_wrapper.handle_async(flask.request.json)

            case utils.ServerType.FASTAPI:
                import fastapi

                @self._server.post(callback_wrapper._endpoint)
                @utils.rename_func(f"('{callback_wrapper._endpoint}')")
                async def pywa_flow(
                    flow_req: EncryptedFlowRequestType,
                ) -> fastapi.Response:
                    """Automatically generated by pywa to handle incoming flow requests."""
                    response, status_code = await callback_wrapper.handle_async(
                        flow_req
                    )
                    return fastapi.Response(
                        content=response,
                        status_code=status_code,
                    )

        return callback_wrapper

    async def _call_handlers(self: "WhatsApp", update: dict) -> None:
        """Call the handlers for the given update."""
        try:
            try:
                handler_type = self._get_handler_type(update)
            except (KeyError, ValueError, TypeError, IndexError):
                (_logger.error if self._validate_updates else _logger.debug)(
                    "Webhook ('%s') received unexpected update%s: %s",
                    self._webhook_endpoint,
                    " (Enable `validate_updates` to ignore updates with invalid data)"
                    if not self._validate_updates
                    else "",
                    update,
                )
                handler_type = None

            if handler_type is None:
                return
            try:
                constructed_update: BaseUpdate = self._handlers_to_updates[
                    handler_type
                ].from_update(client=self, update=update)
                if await self._process_listener(constructed_update):
                    return
                await self._invoke_callbacks(handler_type, constructed_update)
            except Exception:
                _logger.exception("Failed to construct update: %s", update)
        finally:
            # Always call raw update handler last
            await self._call_raw_update_handler(update)

    async def _call_raw_update_handler(self: "WhatsApp", update: dict) -> None:
        """Invoke the raw update handler."""
        await self._invoke_callbacks(RawUpdateHandler, update)

    async def _invoke_callbacks(
        self: "WhatsApp", handler_type: type[Handler], update: BaseUpdate | dict
    ) -> None:
        """Process and call registered handlers for the update."""
        for handler in self._handlers[handler_type]:
            try:
                handled = await handler.ahandle(self, update)
            except StopHandling:
                break
            except ContinueHandling:
                continue
            except Exception:
                handled = True
                _logger.exception(
                    "An error occurred while '%s' was handling an update",
                    handler._callback.__name__,
                )
            if handled and not self._continue_handling:
                break

    async def _process_listener(self: "WhatsApp", update: BaseUpdate) -> bool:
        """Process and answer a listener if present."""
        listener = self._listeners.get(update.listener_identifier)
        if not listener:
            return False

        try:
            if await listener.apply_filters(self, update):
                listener.set_result(update)
                return not self._continue_handling
            elif await listener.apply_cancelers(self, update):
                listener.cancel(update)
                return not self._continue_handling
            else:
                return False  # if no filters or cancelers matched, continue handling
        except ContinueHandling:
            return False
        except StopHandling:
            return True
        except Exception as e:
            listener.set_exception(e)

        return not self._continue_handling

    def _register_callback_url(
        self: "WhatsApp",
        callback_url: str,
        callback_url_scope: utils.CallbackURLScope,
        app_id: int,
        app_secret: str,
        verify_token: str,
        fields: tuple[str, ...] | None,
    ) -> None:
        """
        This is a non-blocking function that registers the callback URL.
        It must be called after the server is running so that the challenge can be verified.
        """
        loop = asyncio.new_event_loop()
        api = copy.copy(self.api)
        api._session = self._httpx_client(  # TODO: copy the session properly
            timeout=api._session.timeout,
            base_url=api._session.base_url,
            headers=api._session.headers,
        )

        try:
            match callback_url_scope:
                case utils.CallbackURLScope.APP:
                    app_access_token = loop.run_until_complete(
                        api.get_app_access_token(app_id=app_id, app_secret=app_secret)
                    )
                    res = loop.run_until_complete(
                        api.set_app_callback_url(
                            app_id=app_id,
                            app_access_token=app_access_token["access_token"],
                            callback_url=callback_url,
                            verify_token=verify_token,
                            fields=fields,
                        )
                    )
                case utils.CallbackURLScope.WABA:
                    res = loop.run_until_complete(
                        api.set_waba_alternate_callback_url(
                            waba_id=self.business_account_id,
                            callback_url=callback_url,
                            verify_token=verify_token,
                        )
                    )
                case utils.CallbackURLScope.PHONE:
                    res = loop.run_until_complete(
                        api.set_phone_alternate_callback_url(
                            callback_url=callback_url,
                            verify_token=verify_token,
                            phone_id=self.phone_id,
                        )
                    )
                case _:
                    raise ValueError("Invalid callback URL scope")

            if not res["success"]:
                raise RuntimeError("Failed to register callback URL.")
            _logger.info("Callback URL '%s' registered successfully", callback_url)
        except errors.WhatsAppError as e:
            raise RuntimeError(
                f"Failed to register callback URL '{callback_url}'. if you are using a slow/custom server, you can "
                "increase the delay using the `webhook_challenge_delay` parameter when initializing the WhatsApp client."
            ) from e



import datetime
import functools
from types import ModuleType

import pytest

from pywa import handlers, types, WhatsApp, filters
from pywa.handlers import FlowRequestHandler, _flow_request_handler_attr
from pywa.types import Message
from pywa_async import WhatsApp as WhatsAppAsync

FAKE_WA = WhatsApp(phone_id="1234567890", token="1234567890:1234567890")
FAKE_WA_ASYNC = WhatsAppAsync(phone_id="1234567890", token="1234567890:1234567890")


CallbackButtonOnlyDataIsNeeded = functools.partial(
    types.CallbackButton,
    _client=FAKE_WA,
    raw={},
    id="123",
    type=types.MessageType.INTERACTIVE,
    metadata=types.Metadata(
        display_phone_number="1234567890",
        phone_number_id="1234567890",
    ),
    from_user=types.User(
        _client=FAKE_WA,
        wa_id="1234567890",
        name="John",
    ),
    timestamp=datetime.datetime.now(datetime.timezone.utc),
    reply_to_message=types.ReplyToMessage(
        message_id="123",
        from_user_id="1234567890",
        referred_product=None,
    ),
    title="Click me",
)


def test_instance_with_parentheses():
    wa = WhatsApp(server=None, verify_token="1234567890")

    @wa.on_message(filters=filters.text)  # @wa.on_x(filters=...)
    def instance_with_parentheses(_, __): ...

    assert (
        wa._handlers[handlers.MessageHandler][0]._callback == instance_with_parentheses
    )


def test_instance_without_parentheses():
    wa = WhatsApp(server=None, verify_token="1234567890")

    @wa.on_message  # @wa.on_x
    def instance_without_parentheses(_, __): ...

    assert (
        wa._handlers[handlers.MessageHandler][0]._callback
        == instance_without_parentheses
    )


def test_class_with_parentheses_kw():
    module = ModuleType("module")

    @WhatsApp.on_message(filters=filters.text)  # @WhatsApp.on_x(filters=...) with kw
    def class_with_parentheses_kw(_, __): ...

    module.__dict__["on_message"] = class_with_parentheses_kw
    wa = WhatsApp(server=None, verify_token="1234567890", handlers_modules=[module])
    assert (
        wa._handlers[handlers.MessageHandler][0]._callback == class_with_parentheses_kw
    )


def test_class_with_parentheses_args():
    module = ModuleType("module")

    @WhatsApp.on_message(filters.text)  # @WhatsApp.on_x(filters) without kw
    def class_with_parentheses_args(_, __): ...

    module.__dict__["on_message"] = class_with_parentheses_args
    wa = WhatsApp(server=None, verify_token="1234567890", handlers_modules=[module])
    assert (
        wa._handlers[handlers.MessageHandler][0]._callback
        == class_with_parentheses_args
    )


def test_class_without_parentheses():
    module = ModuleType("module")

    @WhatsApp.on_message
    def class_without_parentheses(_, __): ...

    module.__dict__["on_message"] = class_without_parentheses
    wa = WhatsApp(server=None, verify_token="1234567890", handlers_modules=[module])
    assert (
        wa._handlers[handlers.MessageHandler][0]._callback == class_without_parentheses
    )


def test_all_combinations():
    wa = WhatsApp(server=None, verify_token="1234567890")

    @wa.on_message(filters=filters.text)
    @wa.on_message
    @WhatsApp.on_message(filters=filters.text)
    @WhatsApp.on_message(filters.text)
    @WhatsApp.on_message
    def all_combinations(_, __): ...

    module = ModuleType("module")
    module.__dict__["on_message"] = all_combinations
    wa.load_handlers_modules(module)
    assert len(wa._handlers[handlers.MessageHandler]) == 5


def test_flow_request_handler():
    @WhatsApp.on_flow_request("/flow")
    def on_flow_class(_, __): ...

    assert isinstance(on_flow_class, handlers.FlowRequestHandler)
    assert hasattr(on_flow_class, _flow_request_handler_attr)

    wa = WhatsApp(server=None, verify_token="1234567890", business_private_key="...")
    on_flow = lambda _, __: ...
    on_flow_instance = wa.get_flow_request_handler(
        endpoint="/flow",
        callback=on_flow,
    )
    assert isinstance(on_flow_instance, handlers.FlowRequestCallbackWrapper)
    assert not hasattr(on_flow_instance, _flow_request_handler_attr)


def test_add_handlers():
    wa = WhatsApp(server=None, verify_token="xyzxyz")
    h = handlers.MessageHandler(
        callback=lambda _, __: None,
        filters=filters.text,
    )
    wa.add_handlers(h)
    assert wa._handlers[handlers.MessageHandler][0] is h


def test_remove_handlers():
    wa = WhatsApp(server=None, verify_token="xyzxyz")
    h = handlers.MessageHandler(
        callback=lambda _, __: None,
        filters=filters.text,
    )
    wa.add_handlers(h)
    wa.remove_handlers(h)
    assert wa._handlers[handlers.MessageHandler] == []

    with pytest.raises(ValueError):
        wa.remove_handlers(h, silent=False)
    wa.remove_handlers(h, silent=True)


def test_remove_callbacks():
    wa = WhatsApp(server=None, verify_token="xyzxyz")
    c = lambda _, __: None
    mh = handlers.MessageHandler(
        callback=c,
        filters=filters.text,
    )
    ch = handlers.CallbackButtonHandler(
        callback=c,
        filters=filters.text,
    )
    wa.add_handlers(mh, ch)
    wa.remove_callbacks(c)
    assert wa._handlers[handlers.MessageHandler] == []
    assert wa._handlers[handlers.CallbackButtonHandler] == []


def test_get_flow_request_handler():
    c = lambda _, __: None
    h = FlowRequestHandler(
        callback=c,
        endpoint="/flow",
    )
    assert h._main_handler is c

    def callback(_, __): ...

    h.add_handler(callback=callback, action=types.FlowRequestActionType.INIT)
    assert h._handlers[(types.FlowRequestActionType.INIT, None)][0][1] is callback


def test_shared_data():
    wa = WhatsApp(server=None, verify_token="xyzxyz")
    msg = Message(
        _client=wa,
        raw={},
        waba_id="456",
        id="123",
        type=types.MessageType.TEXT,
        forwarded=False,
        forwarded_many_times=False,
        reply_to_message=None,
        metadata=types.Metadata(
            display_phone_number="1234567890",
            phone_number_id="1234567890",
        ),
        from_user=types.User(
            _client=wa,
            wa_id="1234567890",
            name="John",
        ),
        timestamp=datetime.datetime.now(datetime.timezone.utc),
    )

    @wa.on_message
    def shared_data_handler(_: WhatsApp, m: types.Message):
        m.shared_data["key"] = None
        m.continue_handling()

    @wa.on_message
    def shared_data_check(_: WhatsApp, m: types.Message):
        m.shared_data["key"] = "value"

    wa._invoke_callbacks(
        handler_type=handlers.MessageHandler,
        update=msg,
    )

    assert msg.shared_data["key"] == "value"
